import warnings

import pandas as pd
import numpy as np
import sys, os
import chaosmagpy as cp
from pathlib import Path
from tqdm import tqdm
from .auxiliaryfunctions import ST_IDW_Process, gg_to_geo

# Get main MagGeo directory (should be parent to this file)
base_dir=os.path.dirname(os.getcwd())
temp_results_dir = os.path.join(base_dir, "temp_data")

TotalSwarmRes_A = pd.read_csv(os.path.join(temp_results_dir,"TotalSwarmRes_A.csv"),low_memory=False, index_col='epoch')
TotalSwarmRes_A['timestamp'] = pd.to_datetime(TotalSwarmRes_A['timestamp'])
TotalSwarmRes_B = pd.read_csv(os.path.join(temp_results_dir,"TotalSwarmRes_B.csv"),low_memory=False, index_col='epoch')
TotalSwarmRes_B['timestamp'] = pd.to_datetime(TotalSwarmRes_B['timestamp'])
TotalSwarmRes_C = pd.read_csv(os.path.join(temp_results_dir,"TotalSwarmRes_C.csv"),low_memory=False, index_col='epoch')
TotalSwarmRes_C['timestamp'] = pd.to_datetime(TotalSwarmRes_C['timestamp'])

def row_handler (GPSData):
    dn = [] ## List used to add all the GPS points with the annotated MAG Data. See the last bullet point of this process        
    for index, row in tqdm(GPSData.iterrows(), total=GPSData.shape[0], desc="Annotating the GPS Trajectory"):
        GPSLat = row['gpsLat']
        GPSLong = row['gpsLong']
        GPSDateTime = row['gpsDateTime']
        GPSTime = row['epoch']
        GPSAltitude = row['gpsAltitude']
        #print("Process for:", index,"Date&Time:",GPSDateTime, "Epoch", GPSTime)
        try:
            result=ST_IDW_Process(GPSLat,GPSLong,GPSAltitude, GPSDateTime,GPSTime, TotalSwarmRes_A, TotalSwarmRes_B, TotalSwarmRes_C)
            dn.append(result)
        except:
            #print("Ups!.That was a bad Swarm Point, let's keep working with the next point")
            result_badPoint= {'Latitude': GPSLat, 'Longitude': GPSLong, 'Altitude':GPSAltitude, 'DateTime': GPSDateTime, 'N_res': np.nan, 'E_res': np.nan, 'C_res':np.nan, 'TotalPoints':0, 'Minimum_Distance':np.nan, 'Average_Distance':np.nan}  
            dn.append(result_badPoint)
            continue

    
    GPS_ResInt = pd.DataFrame(dn)
    GPS_ResInt.to_csv (os.path.join(temp_results_dir,"GPS_ResInt.csv"), header=True)

    def CHAOS_ground_values(GPS_ResInt):
        
        base_dir=os.path.dirname(os.getcwd())
        utilities_dir = os.path.join(base_dir, "utilities")
        
        #1. Load the required parameters, including a local CHAOS model in mat format.
        model = cp.load_CHAOS_matfile(os.path.join(utilities_dir,'CHAOS-7.mat'))
        theta = 90-GPS_ResInt['Latitude'].values
        phi = GPS_ResInt['Longitude'].values
        alt=GPS_ResInt['Altitude'].values
        rad_geoc_ground, theta_geoc_ground, sd_ground, cd_ground = gg_to_geo(alt, theta) # gg_to_geo, will transform the coordinates from geocentric values to geodesic values. Altitude must be in km
        time= cp.data_utils.mjd2000(pd.DatetimeIndex(GPS_ResInt['DateTime']).year, pd.DatetimeIndex(GPS_ResInt['DateTime']).month, pd.DatetimeIndex(GPS_ResInt['DateTime']).day)
        
        #2. Compute the core, crust and magnetosphere contributions at the altitude level.
        B_r_core, B_t_core, B_phi_core = model.synth_values_tdep(time, rad_geoc_ground, theta_geoc_ground, phi) #Core Contribution
        B_r_crust, B_t_crust, B_phi_crust = model.synth_values_static(rad_geoc_ground, theta_geoc_ground, phi) #Crust Contribution
        B_r_magneto, B_t_magneto, B_phi_magneto = model.synth_values_gsm(time, rad_geoc_ground, theta_geoc_ground, phi) #Magnetosphere contribution.

        #3. Change the direction of the axis from XYZ to r,theta and phi.
        B_r_swarm, B_t_swarm, B_phi_swarm = -GPS_ResInt['C_res'], -GPS_ResInt['N_res'], GPS_ResInt['E_res']

        #4. Compute the magnetic component (r,theta,phi) at ground level.
        B_r_ground = B_r_core + B_r_crust + B_r_magneto + B_r_swarm #(-Z)
        B_t_ground = B_t_core + B_t_crust + B_t_magneto + B_t_swarm #(-X)
        B_phi_ground = B_phi_core + B_phi_crust +B_phi_magneto + B_phi_swarm #(Y)

        #4b. Compute the CHAOS internal magnetic component (r,theta,phi) at ground level.
        B_r_ground_internal = B_r_core + B_r_crust #(-Z)
        B_t_ground_internal = B_t_core + B_t_crust #(-X)
        B_phi_ground_internal = B_phi_core + B_phi_crust #(Y)

        #5. Convert B_r_, B_t_, and B_phi to XYZ (NEC)
        Z_chaos = -B_r_ground   #Z
        X_chaos = -B_t_ground   #X
        Y_chaos = B_phi_ground  #Y

        #5b. Convert full field with CHAOS internal only B_r_, B_t_, and B_phi to XYZ (NEC)
        Z_chaos_internal = -B_r_ground_internal #Z
        X_chaos_internal = -B_t_ground_internal #X
        Y_chaos_internal = B_phi_ground_internal #Y

        #6. Rotate the X(N) and Z(C) magnetic field values of the chaos models into the geodetic frame using the sd and cd (sine and cosine d from gg_to_geo) 
        X_obs = X_chaos*cd_ground + Z_chaos*sd_ground #New N
        Z_obs = Z_chaos*cd_ground - X_chaos*sd_ground #New C
        Y_obs = Y_chaos # New E

        #6b. Rotate the X(N) and Z(C) magnetic field values of the internal part of CHAOS models into the geodetic frame using the sd and cd (sine and cosine d from gg_to_geo) 
        X_obs_internal = X_chaos_internal *cd_ground + Z_chaos_internal*sd_ground #New N
        Z_obs_internal = Z_chaos_internal *cd_ground - X_chaos_internal*sd_ground #New C
        Y_obs_internal = Y_chaos_internal # New E

        return X_obs, Y_obs, Z_obs, X_obs_internal, Y_obs_internal, Z_obs_internal

    X_obs, Y_obs, Z_obs, X_obs_internal, Y_obs_internal, Z_obs_internal =CHAOS_ground_values(GPS_ResInt)

    GPS_ResInt['N'] =pd.Series(X_obs)
    GPS_ResInt['E'] =pd.Series(Y_obs)
    GPS_ResInt['C'] =pd.Series(Z_obs)
    GPS_ResInt['N_Obs'] =pd.Series(X_obs_internal)
    GPS_ResInt['E_Obs'] =pd.Series(Y_obs_internal)
    GPS_ResInt['C_Obs'] =pd.Series(Z_obs_internal)

    GPS_ResInt.drop(columns=['N_res', 'E_res','C_res'], inplace=True)

    return GPS_ResInt


"""
Debug utilities for MagGeo package.
"""
import os
import logging
import pandas as pd
from pathlib import Path
from typing import Any, Optional

class MagGeoDebugger:
    """Debug helper for MagGeo operations."""
    
    def __init__(self, debug_mode: Optional[bool] = None):
        # Check if debug mode is enabled via environment variable or parameter
        self.debug_mode = debug_mode if debug_mode is not None else os.getenv('MAGGEO_DEBUG', 'false').lower() == 'true'
        self.temp_dir = Path("temp_data")
        
        if self.debug_mode:
            self._setup_debug_environment()
    
    def _setup_debug_environment(self):
        """Setup debug environment: create temp directory and configure logging."""
        # Ensure temp_data directory exists
        self.temp_dir.mkdir(exist_ok=True)
        
        # Configure logging
        logging.basicConfig(
            level=logging.DEBUG,
            format='[%(asctime)s] %(levelname)s - %(message)s',
            handlers=[
                logging.StreamHandler(),  # Console output
                logging.FileHandler(self.temp_dir / 'maggeo_debug.log')  # File output
            ]
        )
        self.logger = logging.getLogger('MagGeo')
        self.logger.info("Debug mode enabled")
    
    def log(self, message: str, level: str = 'info'):
        """Log a debug message if debug mode is enabled."""
        if self.debug_mode:
            getattr(self.logger, level.lower())(message)
    
    def save_dataframe(self, df: pd.DataFrame, filename: str, description: str = ""):
        """Save DataFrame to temp directory if debug mode is enabled."""
        if self.debug_mode:
            filepath = self.temp_dir / filename
            df.to_csv(filepath, index=False)
            self.log(f"Saved {description}: {filepath} (shape: {df.shape})")
    
    def print_dataframe_info(self, df: pd.DataFrame, name: str):
        """Print DataFrame info if debug mode is enabled."""
        if self.debug_mode:
            self.log(f"{name} DataFrame shape: {df.shape}")
            self.log(f"{name} head:\n{df.head()}")
    
    def log_parameters(self, params: dict):
        """Log parameters if debug mode is enabled."""
        if self.debug_mode:
            self.log(f"Parameters: {params}")
    
    def log_unique_dates(self, dates):
        """Log unique dates if debug mode is enabled."""
        if self.debug_mode:
            self.log(f"Unique dates: {list(dates)}")
    
    def log_date_range(self, date, start_date, end_date):
        """Log date range for Swarm data request."""
        if self.debug_mode:
            self.log(f"Requesting data for {date}: {start_date} to {end_date}")
    
    def log_swarm_data(self, satellite: str, date, df: pd.DataFrame):
        """Log Swarm data information."""
        if self.debug_mode:
            self.log(f"Swarm {satellite} DataFrame for {date}: shape {df.shape}")
            self.save_dataframe(df, f"swarm_{satellite}_{date}.csv", f"Swarm {satellite} data for {date}")
    
    def log_swarm_concat(self, satellite: str, df: pd.DataFrame):
        """Log concatenated Swarm data information."""
        if self.debug_mode:
            self.log(f"Swarm {satellite} Concat DataFrame: shape {df.shape}")
            self.save_dataframe(df, f"swarm_Concat_{satellite}.csv", f"Concatenated Swarm {satellite} data")
    
    def log_chaos_values(self, component: str, values):
        """Log CHAOS model values.""" 
        if self.debug_mode:
            self.log(f"{component} values: {values}")
    
    def log_interpolation_result(self, df: pd.DataFrame):
        """Log interpolation results."""
        if self.debug_mode:
            self.log(f"After interpolation, DataFrame shape: {df.shape}")
            self.save_dataframe(df, "interpolated.csv", "Interpolated data")
    
    def log_chaos_result(self, df: pd.DataFrame):
        """Log CHAOS model results."""
        if self.debug_mode:
            self.log(f"After CHAOS model, DataFrame shape: {df.shape}")
            self.save_dataframe(df, "chaos_annotated.csv", "CHAOS annotated data")

    def save_maggeo_results(self, df: pd.DataFrame, filename: str, description: str = ""):
            """Save MagGeo Results to temp directory if debug mode is enabled."""
            if self.debug_mode:
                filepath = self.temp_dir / filename
                df.to_csv(filepath, index=False)
                self.log(f"Saved {description}: {filepath} (shape: {df.shape})")


# Global debugger instance
_debugger = None

def get_debugger() -> MagGeoDebugger:
    """Get or create the global debugger instance."""
    global _debugger
    if _debugger is None:
        _debugger = MagGeoDebugger()
    return _debugger



"""
Parallel processing utilities for MagGeo interpolation and CHAOS calculations.

This module provides functions to parallelize the MagGeo pipeline correctly:
- Only GPS trajectory data is chunked for parallel processing
- Complete Swarm data (A, B, C) is passed to each worker process
- Each GPS point can find its matches across all Swarm data for proper interpolation
- CHAOS calculations are performed after interpolation with correct data flow
"""

import multiprocessing as mp
import pandas as pd
import numpy as np
from typing import List, Tuple, Dict, Any, Optional
from functools import partial
import os
from tqdm import tqdm
from .interpolation import st_idw_process
from .chaos import chaos_ground_values


def get_optimal_chunk_size(total_gps_points: int, n_cores: int, min_chunk_size: int = 50) -> int:
    """
    Calculate optimal chunk size for GPS trajectory parallel processing.
    
    Parameters
    ----------
    total_gps_points : int
        Total number of GPS points in the trajectory
    n_cores : int
        Number of CPU cores available
    min_chunk_size : int, default 50
        Minimum chunk size to ensure efficiency
        
    Returns
    -------
    int
        Optimal chunk size for GPS trajectory chunking
    """
    # Create chunks that will keep workers busy but not too small to cause overhead
    chunk_size = max(total_gps_points // (n_cores * 4), min_chunk_size)
    return min(chunk_size, 200)  # Cap at 200 to avoid memory issues


def split_gps_trajectory_into_chunks(gps_df: pd.DataFrame, chunk_size: int) -> List[pd.DataFrame]:
    """
    Split GPS trajectory DataFrame into chunks for parallel processing.
    
    IMPORTANT: Only the GPS trajectory is chunked. Swarm data must remain complete
    for each worker to find proper matches for interpolation.
    
    Parameters
    ----------
    gps_df : pd.DataFrame
        GPS trajectory DataFrame to split
    chunk_size : int
        Size of each GPS chunk
        
    Returns
    -------
    List[pd.DataFrame]
        List of GPS trajectory chunks
    """
    chunks = []
    for i in range(0, len(gps_df), chunk_size):
        chunk = gps_df.iloc[i:i + chunk_size].copy()
        chunks.append(chunk)
    return chunks


def process_gps_chunk_complete_pipeline(
    chunk_data: Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, int]
) -> pd.DataFrame:
    """
    Process a GPS trajectory chunk through the complete MagGeo pipeline.
    
    This function follows the correct MagGeo logic:
    1. For each GPS point in the chunk, interpolate using ALL Swarm data
    2. Calculate CHAOS ground values for all interpolated points
    3. Calculate additional geomagnetic components
    
    Parameters
    ----------
    chunk_data : tuple
        Tuple containing (gps_chunk, complete_swarm_a, complete_swarm_b, complete_swarm_c, dt_seconds)
        
    Returns
    -------
    pd.DataFrame
        Complete annotated DataFrame chunk with all geomagnetic values
    """
    gps_chunk, swarm_a_complete, swarm_b_complete, swarm_c_complete, dt_seconds = chunk_data
    
    # Step 1: Interpolation phase - process each GPS point with access to ALL Swarm data
    interpolation_results = []
    for _, gps_row in gps_chunk.iterrows():
        try:
            # Use the complete Swarm datasets for proper spatiotemporal matching
            interpolation_result = st_idw_process(
                gps_row['gpsLat'], gps_row['gpsLong'], gps_row['gpsAltitude'],
                gps_row['gpsDateTime'], gps_row['epoch'],
                swarm_a_complete, swarm_b_complete, swarm_c_complete, dt_seconds
            )
            interpolation_results.append(interpolation_result)
        except Exception as e:
            # Bad Point could be not points in Swarm data or other issues.
            bad_point_result = {
                'Latitude': gps_row['gpsLat'],
                'Longitude': gps_row['gpsLong'],
                'Altitude': gps_row['gpsAltitude'],
                'DateTime': gps_row['gpsDateTime'],
                'N_res': float('nan'),
                'E_res': float('nan'),
                'C_res': float('nan'),
                'TotalPoints': 0,
                'Minimum_Distance': float('nan'),
                'Average_Distance': float('nan'),
                'Kp': float('nan')
            }
            interpolation_results.append(bad_point_result)
    
    # Convert interpolation results to DataFrame
    gps_with_residuals = pd.DataFrame(interpolation_results)
    
    # Step 2: CHAOS calculations phase - only after interpolation is complete
    try:
        # Calculate CHAOS ground values for the entire chunk
        X_obs, Y_obs, Z_obs, X_obs_internal, Y_obs_internal, Z_obs_internal = chaos_ground_values(gps_with_residuals)
        
        # Add CHAOS results to the DataFrame
        gps_with_residuals['N'] = X_obs
        gps_with_residuals['E'] = Y_obs
        gps_with_residuals['C'] = Z_obs
        gps_with_residuals['N_Obs'] = X_obs_internal
        gps_with_residuals['E_Obs'] = Y_obs_internal
        gps_with_residuals['C_Obs'] = Z_obs_internal
        
        # Step 3: Calculate additional geomagnetic components
        gps_with_residuals['H'] = np.sqrt((gps_with_residuals['N']**2 + gps_with_residuals['E']**2))
        gps_with_residuals['D'] = np.degrees(np.arctan2(gps_with_residuals['E'], gps_with_residuals['N']))
        gps_with_residuals['I'] = np.degrees(np.arctan2(gps_with_residuals['C'], gps_with_residuals['H']))
        gps_with_residuals['F'] = np.sqrt((gps_with_residuals['N']**2 + gps_with_residuals['E']**2 + gps_with_residuals['C']**2))
        
    except Exception as e:
        # If CHAOS calculation fails, add NaN columns, this need to be handled gracefully
        for col in ['N', 'E', 'C', 'N_Obs', 'E_Obs', 'C_Obs', 'H', 'D', 'I', 'F']:
            gps_with_residuals[col] = float('nan')
    
    return gps_with_residuals


def parallel_maggeo_annotation(
    gps_df: pd.DataFrame,
    swarm_a: pd.DataFrame,
    swarm_b: pd.DataFrame,
    swarm_c: pd.DataFrame,
    dt_seconds: int = 14400,
    n_cores: Optional[int] = None,
    chunk_size: Optional[int] = None
) -> pd.DataFrame:
    
    """
    Perform parallel MagGeo annotation following the correct logic.
   
    - Only GPS trajectory is chunked for parallel processing
    - Complete Swarm data (A, B, C) is passed to each core
    - Each GPS point finds matches across ALL Swarm data for proper interpolation.
    - this could be enhaced with a better approach in the future.
    - CHAOS calculations follow after interpolation with correct data flow
    
    Parameters
    ----------
    gps_df : pd.DataFrame
        GPS trajectory DataFrame
    swarm_a, swarm_b, swarm_c : pd.DataFrame
        Complete Swarm satellite data DataFrames (NOT chunked)
    dt_seconds : int, default 14400
        Time window in seconds for interpolation
    n_cores : int, optional
        Number of cores to use. If None, uses all available cores.
    chunk_size : int, optional
        Size of GPS chunks for processing. If None, calculated automatically.
        
    Returns
    -------
    pd.DataFrame
        Complete annotated DataFrame with all geomagnetic values
    """
    #TODO: Adjust this to use when parallel is active and validate why it runs twice
    #print(f"🚀 Starting parallel MagGeo annotation:")
    #print(f"   📍 GPS points: {len(gps_df)}")
    #print(f"   🛰️ Swarm A records: {len(swarm_a)}")
    #print(f"   🛰️ Swarm B records: {len(swarm_b)}")
    #print(f"   🛰️ Swarm C records: {len(swarm_c)}")
    #print(f"   🔧 Using {n_cores} cores with chunk size: {chunk_size}")

    if n_cores is None:
        n_cores = mp.cpu_count()
    
    if chunk_size is None:
        chunk_size = get_optimal_chunk_size(len(gps_df), n_cores)
    
    # Split ONLY the GPS trajectory into chunks (Swarm data remains complete)
    gps_chunks = split_gps_trajectory_into_chunks(gps_df, chunk_size)
    
    # Prepare data for multiprocessing - each worker gets complete Swarm data
    chunk_data_list = [
        (gps_chunk, swarm_a, swarm_b, swarm_c, dt_seconds) 
        for gps_chunk in gps_chunks
    ]
    
    #print(f"📦 Created {len(gps_chunks)} GPS chunks for parallel processing")
    
    # Process GPS chunks in parallel with complete pipeline
    with mp.Pool(processes=n_cores) as pool:
        results = list(tqdm(
            pool.imap(process_gps_chunk_complete_pipeline, chunk_data_list),
            total=len(chunk_data_list),
            desc="🧮 Processing GPS chunks (interpolation + CHAOS)",
            unit="chunk"
        ))
    
    # Concatenate all results
    final_df = pd.concat(results, ignore_index=True)
    
    # Clean up intermediate columns
    columns_to_drop = ['N_res', 'E_res', 'C_res']
    final_df.drop(columns=[col for col in columns_to_drop if col in final_df.columns], inplace=True)
    
    print(f"✅ Parallel annotation completed: {len(final_df)} points processed")
    return final_df


import numpy as np
import pandas as pd

def distance_to_GPS(s_lat, s_lng, e_lat, e_lng):
    """ Calculate the distance between two GPS coordinates.

    This function uses the Haversine formula to compute the distance between two points on the Earth specified by their latitude and longitude.
    It returns the distance in kilometers.
    
    Parameters
    ----------
    s_lat : float
        Starting latitude in degrees.
    s_lng : float
        Starting longitude in degrees.
    e_lat : float
        Ending latitude in degrees.
    e_lng : float
        Ending longitude in degrees.
    Returns
    -------
    float
        Distance in kilometers between the two GPS coordinates.
    """
    # Haversine formula to calculate the distance between two points on the Earth
    # Reference: https://en.wikipedia.org/wiki/Haversine_formula
    
    R = 6373.0
    # approximate radius of earth in km
    s_lat = s_lat*(np.pi)/180.0                     
    s_lng = np.deg2rad(s_lng)     
    e_lat = np.deg2rad(e_lat)                       
    e_lng = np.deg2rad(e_lng)
    d = np.sin((e_lat - s_lat)/2)**2 + np.cos(s_lat)*np.cos(e_lat) * np.sin((e_lng - s_lng)/2)**2
    return 2 * R * np.arcsin(np.sqrt(d))

def Kradius(lat):
    """
    Calculate the radius of the Earth at a given latitude.

    Parameters
    ----------
    lat : float
        Latitude in degrees. Must be in the range (-90, 90), exclusive.

    Returns
    -------
    float
        Radius in kilometers at the given latitude.

    Raises
    ------
    ValueError
        If latitude is not in the range (-90, 90).
    """
    if 0 <= lat < 90:
        # for Northern Latitudes 
        nlat = (-10 * lat) + 1800
        #print("The R on the North")
    elif -90 < lat < 0:
        # for Southern Latitudes
        nlat = (10 * lat) + 1800
        #print ("The R on the South")
    else:
        raise ValueError("Latitude must be between -90 and 90 degrees (exclusive).")
    return nlat


def DistJ(ds, r, dt, DT):
    """
    Calculate the distance between two points in a geodetic frame
    See the paper "Fusion of wildlife tracking and satellite geomagnetic data for the study of animal migration" by Benitez-Paez et al. (2023) for more details.
    Parameters
    ----------
    ds : float
        Distance in the North-South direction.
    r : float
        Radius of the Earth at the given latitude.
    dt : float
        Distance in the East-West direction.
    DT : float
        Time difference in seconds.
    
    Returns
    -------
    float
        Distance in kilometers between the two points.
    """
    eDist = np.sqrt(((ds/r)**2 + (dt/DT)**2)/2)
    return eDist

def DfTime_func(SwarmData, GPSTime, DT):
    """
    Extract a DataFrame for a specific GPS time and its surrounding data.

    Parameters
    ----------
    SwarmData : pd.DataFrame
        DataFrame containing Swarm data indexed by time.
    GPSTime : pd.Timestamp
        The GPS time for which to extract the DataFrame in Epoch format.
    DT : pd.Timedelta
        Time delta to define the range around GPSTime.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the data for the specified GPS time and its surrounding data.
        If `GPSTime` is not found in the index, returns an empty DataFrame.
    """
    # Check if GPSTime is in the index of SwarmData
    import logging
    logging.debug("Entered DfTime_func")
    if GPSTime in SwarmData.index:
        logging.debug(f"Extracting data for GPSTime: {GPSTime} with DT: {DT}")
        # Extract the DataFrame for the specified GPS time and its surrounding data
        logging.debug(f"Yes the GPSTime is in the index of SwarmData:{GPSTime}, should be the same")
        return SwarmData.loc[GPSTime-DT:GPSTime+DT]
    else:
        logging.debug(f"GPSTime {GPSTime} not found in SwarmData index. Returning empty DataFrame.")
        return pd.DataFrame()
    
def gg_to_geo(h, gdcolat):
    """
Compute geocentric colatitude and radius from geodetic colatitude and height.

Parameters
----------
h : ndarray, shape (...) Altitude in kilometers.

gdcolat : ndarray, shape (...) Geodetic colatitude

Returns
-------
radius : ndarray, shape (...) Geocentric radius in kilometers.

theta : ndarray, shape (...) Geocentric colatitude in degrees.

sd : ndarray shape (...)  rotate B_X to gd_lat 

cd : ndarray shape (...) rotate B_Z to gd_lat 

References
----------

* Equations (51)-(53) from "The main field" (chapter 4) by Langel, R. A. in:

"Geomagnetism", Volume 1, Jacobs, J. A., Academic Press, 1987.

* Malin, S.R.C. and Barraclough, D.R., 1981. An algorithm for synthesizing 

the geomagnetic field. Computers & Geosciences, 7(4), pp.401-405.

"""

# Use WGS-84 ellipsoid parameters

    eqrad = 6378.137 # equatorial radius

    flat = 1/298.257223563 

    plrad = eqrad*(1-flat) # polar radius

    ctgd = np.cos(np.deg2rad(gdcolat))

    stgd = np.sin(np.deg2rad(gdcolat))

    a2 = eqrad*eqrad

    a4 = a2*a2

    b2 = plrad*plrad

    b4 = b2*b2

    c2 = ctgd*ctgd

    s2 = 1-c2

    rho = np.sqrt(a2*s2 + b2*c2)

    rad = np.sqrt(h*(h+2*rho) + (a4*s2+b4*c2)/rho**2)

    cd = (h+rho)/rad

    sd = (a2-b2)*ctgd*stgd/(rho*rad)

    cthc = ctgd*cd - stgd*sd # Also: sthc = stgd*cd + ctgd*sd

    thc = np.rad2deg(np.arccos(cthc)) # arccos returns values in [0, pi]

    return rad, thc, sd, cd


"""
Swarm Data Manager for MagGeo

This module provides independent functionality for downloading, saving, and loading
Swarm satellite data. It can be used independently from the main MagGeo pipeline
for users who want to:
1. Download Swarm data from large trajectories and save locally
2. Process data in batches 
3. Reuse previously downloaded data
4. Run interpolation processes separately

Key features:
- Batch downloading with progress tracking
- Multiple save/load formats (CSV, Parquet*)
- Data validation and quality checks
- Resume capability for interrupted downloads
- Memory-efficient processing for large datasets
"""

import os
import datetime as dt
from pathlib import Path
from typing import Dict, Any, List, Tuple, Optional, Union
import pandas as pd
from tqdm import tqdm
import warnings

from .swarm import get_swarm_residuals
from .date_utils import identify_unique_dates
from .debug import get_debugger


class SwarmDataManager:
    """
    Manages Swarm satellite data download, storage, and retrieval operations.
    
    This class provides a high-level interface for working with Swarm data
    independently from the main MagGeo pipeline.
    """
    
    def __init__(self, 
                 data_dir: str = "swarm_data",
                 file_format: str = "csv",
                 chunk_size: int = 10,
                 token: Optional[str] = None):
        """
        Initialize SwarmDataManager.
        
        Parameters
        ----------
        data_dir : str, default "swarm_data"
            Directory to store downloaded Swarm data
        file_format : str, default "parquet" 
            File format for saving data. Options: "csv", "parquet"
        chunk_size : int, default 10
            Number of dates to process in each batch
        token : str, optional
            VirES token for authentication
        """
        self.data_dir = Path(data_dir)
        self.data_dir.mkdir(exist_ok=True)
        
        self.file_format = file_format.lower()
        if self.file_format not in ["csv", "parquet"]:
            raise ValueError("file_format must be one of: csv, or parquet")
            
        self.chunk_size = chunk_size
        self.token = token
        self.debugger = get_debugger()
        
        # Create subdirectories for each satellite
        for sat in ['A', 'B', 'C']:
            (self.data_dir / f"swarm_{sat}").mkdir(exist_ok=True)
    
    def download_for_trajectory(self,
                              gps_df: pd.DataFrame,
                              save_individual_files: bool = True,
                              save_concatenated: bool = True,
                              resume: bool = True) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Download Swarm data for an entire GPS trajectory.
        
        Parameters
        ----------
        gps_df : pd.DataFrame
            GPS trajectory data with datetime information
        save_individual_files : bool, default True
            Whether to save individual daily files
        save_concatenated : bool, default True
            Whether to save concatenated files for each satellite
        resume : bool, default True
            Whether to skip already downloaded files
            
        Returns
        -------
        tuple
            Tuple of concatenated DataFrames for satellites A, B, C
        """
        self.debugger.log("Starting Swarm data download for trajectory")
        
        # Identify unique dates
        unique_dates_df = identify_unique_dates(gps_df)
        unique_dates = unique_dates_df['date']
        
        self.debugger.log(f"Need to download data for {len(unique_dates)} unique dates")
        
        return self.download_for_dates(
            unique_dates,
            save_individual_files=save_individual_files,
            save_concatenated=save_concatenated,
            resume=resume
        )
    
    def download_for_dates(self,
                          dates: List[dt.date],
                          save_individual_files: bool = True,
                          save_concatenated: bool = True,
                          resume: bool = True) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Download Swarm data for specific dates.
        
        Parameters
        ----------
        dates : List[dt.date]
            List of dates to download data for
        save_individual_files : bool, default True
            Whether to save individual daily files
        save_concatenated : bool, default True
            Whether to save concatenated files for each satellite
        resume : bool, default True
            Whether to skip already downloaded files
            
        Returns
        -------
        tuple
            Tuple of concatenated DataFrames for satellites A, B, C
        """
        swarm_a_list, swarm_b_list, swarm_c_list = [], [], []
        
        # Filter dates if resuming
        if resume:
            dates_to_process = self._filter_existing_dates(dates)
            if len(dates_to_process) < len(dates):
                self.debugger.log(f"Resuming: {len(dates) - len(dates_to_process)} dates already downloaded")
                print(f"📁 Found existing data for {len(dates) - len(dates_to_process)} dates")
        else:
            dates_to_process = dates
        
        if not dates_to_process:
            print("✅ All Swarm data already downloaded!")
            # Load existing data
            existing_data = self._load_existing_data(dates)
            if existing_data:
                swarm_a_concat = pd.concat(existing_data[0]) if existing_data[0] else pd.DataFrame()
                swarm_b_concat = pd.concat(existing_data[1]) if existing_data[1] else pd.DataFrame()
                swarm_c_concat = pd.concat(existing_data[2]) if existing_data[2] else pd.DataFrame()
                return swarm_a_concat, swarm_b_concat, swarm_c_concat
        
        # Download data for each date with unified progress bar
        hours_added = dt.timedelta(hours=23, minutes=59, seconds=59)
        
        print(f"🛰️ Downloading Swarm data for {len(dates_to_process)} dates...")
        pbar = tqdm(dates_to_process, 
                   desc="Swarm Data Progress", 
                   unit="day",
                   bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} days [{elapsed}<{remaining}]')
        
        for date in pbar:
            try:
                # Update progress bar description with current date
                pbar.set_description(f"Downloading {date.strftime('%Y-%m-%d')}")
                
                startdate = dt.datetime.combine(date, dt.datetime.min.time())
                enddate = startdate + hours_added
                
                self.debugger.log(f"Downloading data for {date}")
                
                # Download data
                swarm_a, swarm_b, swarm_c = get_swarm_residuals(
                    startdate, enddate, self.token
                )
                
                # Add date metadata
                for df, sat_name in [(swarm_a, 'A'), (swarm_b, 'B'), (swarm_c, 'C')]:
                    df['download_date'] = date
                    df['data_quality'] = self._assess_data_quality(df)
                
                # Save individual files if requested
                if save_individual_files:
                    self._save_daily_data(date, swarm_a, swarm_b, swarm_c)
                
                swarm_a_list.append(swarm_a)
                swarm_b_list.append(swarm_b)
                swarm_c_list.append(swarm_c)
                
                # Update progress bar with success
                pbar.set_postfix(status="✓", refresh=False)
                
            except Exception as e:
                self.debugger.log(f"Error downloading data for {date}: {str(e)}", "error")
                warnings.warn(f"Failed to download data for {date}: {str(e)}")
                # Update progress bar with error
                pbar.set_postfix(status="✗", refresh=False)
                continue
        
        pbar.close()
        print("🎯 Swarm data download completed!")
        
        # Load any existing data if resuming
        if resume:
            existing_data = self._load_existing_data(dates)
            if existing_data:
                swarm_a_list.extend(existing_data[0])
                swarm_b_list.extend(existing_data[1]) 
                swarm_c_list.extend(existing_data[2])
        
        # Concatenate all data
        swarm_a_concat = pd.concat(swarm_a_list) if swarm_a_list else pd.DataFrame()
        swarm_b_concat = pd.concat(swarm_b_list) if swarm_b_list else pd.DataFrame()
        swarm_c_concat = pd.concat(swarm_c_list) if swarm_c_list else pd.DataFrame()
        
        # Save concatenated files if requested
        if save_concatenated and not swarm_a_concat.empty:
            self._save_concatenated_data(swarm_a_concat, swarm_b_concat, swarm_c_concat)
        
        self.debugger.log("Swarm data download completed")
        return swarm_a_concat, swarm_b_concat, swarm_c_concat
    
    def load_data_for_dates(self, 
                           dates: List[dt.date],
                           satellites: List[str] = ['A', 'B', 'C']) -> Dict[str, pd.DataFrame]:
        """
        Load previously downloaded Swarm data for specific dates.
        
        Parameters
        ----------
        dates : List[dt.date]
            List of dates to load data for
        satellites : List[str], default ['A', 'B', 'C']
            Which satellites to load data for
            
        Returns
        -------
        dict
            Dictionary with satellite names as keys and concatenated DataFrames as values
        """
        result = {}
        
        for sat in satellites:
            sat_data = []
            for date in dates:
                try:
                    daily_data = self._load_daily_data(date, sat)
                    if daily_data is not None:
                        sat_data.append(daily_data)
                except Exception as e:
                    self.debugger.log(f"Could not load data for {sat} on {date}: {str(e)}", "warning")
            
            if sat_data:
                result[sat] = pd.concat(sat_data)
            else:
                result[sat] = pd.DataFrame()
                
        return result
    
    def load_concatenated_data(self, 
                             satellites: List[str] = ['A', 'B', 'C']) -> Dict[str, pd.DataFrame]:
        """
        Load previously saved concatenated Swarm data.
        
        Parameters
        ---------- 
        satellites : List[str], default ['A', 'B', 'C']
            Which satellites to load data for
            
        Returns
        -------
        dict
            Dictionary with satellite names as keys and DataFrames as values
        """
        result = {}
        
        for sat in satellites:
            try:
                filepath = self.data_dir / f"swarm_{sat}_concatenated.{self.file_format}"
                
                if filepath.exists():
                    result[sat] = self._load_file(filepath)
                    self.debugger.log(f"Loaded concatenated data for satellite {sat}: {result[sat].shape}")
                else:
                    self.debugger.log(f"No concatenated data found for satellite {sat}", "warning")
                    result[sat] = pd.DataFrame()
                    
            except Exception as e:
                self.debugger.log(f"Error loading concatenated data for satellite {sat}: {str(e)}", "error")
                result[sat] = pd.DataFrame()
                
        return result
    
    def get_data_summary(self) -> pd.DataFrame:
        """
        Get summary of available downloaded data.
        
        Returns
        -------
        pd.DataFrame
            Summary of available data files with metadata
        """
        summary_data = []
        
        for sat in ['A', 'B', 'C']:
            sat_dir = self.data_dir / f"swarm_{sat}"
            
            if sat_dir.exists():
                for file_path in sat_dir.glob(f"*.{self.file_format}"):
                    # Extract date from filename
                    date_str = file_path.stem.replace(f"swarm_{sat}_", "")
                    
                    try:
                        file_date = dt.datetime.strptime(date_str, "%Y-%m-%d").date()
                        file_size = file_path.stat().st_size
                        
                        # Load file to get row count
                        df = self._load_file(file_path)
                        
                        summary_data.append({
                            'satellite': sat,
                            'date': file_date,
                            'filename': file_path.name,
                            'file_size_mb': round(file_size / (1024 * 1024), 2),
                            'row_count': len(df),
                            'data_quality': df['data_quality'].iloc[0] if 'data_quality' in df.columns else 'unknown'
                        })
                        
                    except Exception as e:
                        self.debugger.log(f"Error processing file {file_path}: {str(e)}", "warning")
        
        return pd.DataFrame(summary_data)
    
    def cleanup_data(self, 
                    older_than_days: Optional[int] = None,
                    quality_threshold: str = 'poor') -> int:
        """
        Clean up downloaded data files.
        
        Parameters
        ----------
        older_than_days : int, optional
            Remove files older than this many days
        quality_threshold : str, default 'poor'
            Remove files with data quality below this threshold
            
        Returns
        -------
        int
            Number of files removed
        """
        files_removed = 0
        
        for sat in ['A', 'B', 'C']:
            sat_dir = self.data_dir / f"swarm_{sat}"
            
            if sat_dir.exists():
                for file_path in sat_dir.glob(f"*.{self.file_format}"):
                    should_remove = False
                    
                    # Check age
                    if older_than_days:
                        file_age = dt.datetime.now() - dt.datetime.fromtimestamp(file_path.stat().st_mtime)
                        if file_age.days > older_than_days:
                            should_remove = True
                    
                    # Check quality
                    if quality_threshold and not should_remove:
                        try:
                            df = self._load_file(file_path)
                            if 'data_quality' in df.columns:
                                quality = df['data_quality'].iloc[0]
                                if quality == 'poor' and quality_threshold in ['good', 'fair']:
                                    should_remove = True
                                elif quality in ['poor', 'fair'] and quality_threshold == 'good':
                                    should_remove = True
                        except:
                            pass
                    
                    if should_remove:
                        file_path.unlink()
                        files_removed += 1
                        self.debugger.log(f"Removed file: {file_path}")
        
        return files_removed
    
    # Private methods
    def _filter_existing_dates(self, dates: List[dt.date]) -> List[dt.date]:
        """Filter out dates that already have downloaded data."""
        dates_to_process = []
        
        for date in dates:
            # Check if all satellites have data for this date
            all_exist = True
            for sat in ['A', 'B', 'C']:
                filepath = self.data_dir / f"swarm_{sat}" / f"swarm_{sat}_{date.strftime('%Y-%m-%d')}.{self.file_format}"
                if not filepath.exists():
                    all_exist = False
                    break
            
            if not all_exist:
                dates_to_process.append(date)
                
        return dates_to_process
    
    def _load_existing_data(self, dates: List[dt.date]) -> Optional[Tuple[List, List, List]]:
        """Load existing data for dates that have already been downloaded."""
        swarm_a_list, swarm_b_list, swarm_c_list = [], [], []
        
        for date in dates:
            try:
                # Check if data exists for all satellites
                data_exists = True
                for sat in ['A', 'B', 'C']:
                    filepath = self.data_dir / f"swarm_{sat}" / f"swarm_{sat}_{date.strftime('%Y-%m-%d')}.{self.file_format}"
                    if not filepath.exists():
                        data_exists = False
                        break
                
                if data_exists:
                    swarm_a = self._load_daily_data(date, 'A')
                    swarm_b = self._load_daily_data(date, 'B') 
                    swarm_c = self._load_daily_data(date, 'C')
                    
                    if swarm_a is not None and swarm_b is not None and swarm_c is not None:
                        swarm_a_list.append(swarm_a)
                        swarm_b_list.append(swarm_b)
                        swarm_c_list.append(swarm_c)
                        
            except Exception as e:
                self.debugger.log(f"Error loading existing data for {date}: {str(e)}", "warning")
        
        if swarm_a_list:
            return swarm_a_list, swarm_b_list, swarm_c_list
        return None
    
    def _save_daily_data(self, date: dt.date, swarm_a: pd.DataFrame, swarm_b: pd.DataFrame, swarm_c: pd.DataFrame):
        """Save daily data for each satellite."""
        date_str = date.strftime('%Y-%m-%d')
        
        for df, sat in [(swarm_a, 'A'), (swarm_b, 'B'), (swarm_c, 'C')]:
            filepath = self.data_dir / f"swarm_{sat}" / f"swarm_{sat}_{date_str}.{self.file_format}"
            self._save_file(df, filepath)
    
    def _load_daily_data(self, date: dt.date, satellite: str) -> Optional[pd.DataFrame]:
        """Load daily data for a specific satellite."""
        date_str = date.strftime('%Y-%m-%d')
        filepath = self.data_dir / f"swarm_{satellite}" / f"swarm_{satellite}_{date_str}.{self.file_format}"
        
        if filepath.exists():
            return self._load_file(filepath)
        return None
    
    def _save_concatenated_data(self, swarm_a: pd.DataFrame, swarm_b: pd.DataFrame, swarm_c: pd.DataFrame):
        """Save concatenated data for each satellite."""
        for df, sat in [(swarm_a, 'A'), (swarm_b, 'B'), (swarm_c, 'C')]:
            filepath = self.data_dir / f"swarm_{sat}_concatenated.{self.file_format}"
            self._save_file(df, filepath)
            self.debugger.log(f"Saved concatenated data for satellite {sat}: {filepath}")
    
    def _save_file(self, df: pd.DataFrame, filepath: Path):
        """Save DataFrame using the specified file format."""
        if self.file_format == "csv":
            df.to_csv(filepath, index=True)
        elif self.file_format == "parquet":
            df.to_parquet(filepath, index=True)
        #elif self.file_format == "pickle":
            #df.to_pickle(filepath)
    
    def _load_file(self, filepath: Path) -> pd.DataFrame:
        """Load DataFrame using the specified file format.""" 
        if self.file_format == "csv":
            return pd.read_csv(filepath, index_col=0)
        elif self.file_format == "parquet":
            return pd.read_parquet(filepath)
    
    def _assess_data_quality(self, df: pd.DataFrame) -> str:
        """Assess the quality of downloaded Swarm data."""
        if df.empty:
            return 'poor'
        
        # Check for missing values in key columns
        key_columns = ['N_res', 'E_res', 'C_res', 'F_res']
        missing_ratio = df[key_columns].isnull().sum().sum() / (len(df) * len(key_columns))
        
        if missing_ratio > 0.5:
            return 'poor'
        elif missing_ratio > 0.1:
            return 'fair'
        else:
            return 'good'


# Convenience functions for backward compatibility and ease of use
def download_swarm_data_for_trajectory(gps_df: pd.DataFrame,
                                      data_dir: str = "swarm_data",
                                      file_format: str = "csv",
                                      token: Optional[str] = None,
                                      resume: bool = True) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Convenience function to download Swarm data for a GPS trajectory.
    
    Parameters
    ----------
    gps_df : pd.DataFrame
        GPS trajectory data
    data_dir : str, default "swarm_data"
        Directory to store data
    file_format : str, default "csv"
        File format for saving data
    token : str, optional
        VirES authentication token
    resume : bool, default True
        Whether to resume from existing downloads
        
    Returns
    -------
    tuple
        Tuple of concatenated DataFrames for satellites A, B, C
    """
    manager = SwarmDataManager(data_dir=data_dir, file_format=file_format, token=token)
    return manager.download_for_trajectory(gps_df, resume=resume)


def load_swarm_data(data_dir: str = "swarm_data",
                   file_format: str = "csv",
                   satellites: List[str] = ['A', 'B', 'C']) -> Dict[str, pd.DataFrame]:
    """
    Convenience function to load previously downloaded Swarm data.
    
    Parameters
    ----------
    data_dir : str, default "swarm_data"
        Directory containing the data
    file_format : str, default "csv"
        File format of the data
    satellites : List[str], default ['A', 'B', 'C']
        Which satellites to load
        
    Returns
    -------
    dict
        Dictionary with satellite names as keys and DataFrames as values
    """
    manager = SwarmDataManager(data_dir=data_dir, file_format=file_format)
    return manager.load_concatenated_data(satellites=satellites)



# 1. For each day in the trajectory, Get the Swarm Data and Residuals: Get_Swarm_and_residuals
# Input:  Date and Time variables
# Output: Swarm DF for each Sat, including the residuals, and Quality Flags.

import os
from viresclient import SwarmRequest

# Global variable to track if token has been set to avoid multiple "Token saved" messages
_token_set = False

def get_swarm_residuals(startDateTime, endDateTime, token=None) -> tuple:
    """Request Swarm data and residuals for a specified time range.
    This function uses the Viresclient to request data from the Swarm mission for three satellites (A, B, C).

    Parameters
    ----------
    startDateTime : pd.Timestamp
        Start date and time for the data request.
    endDateTime : pd.Timestamp
        End date and time for the data request.
    token : str, optional
        VirES token for authentication. If provided, will be set before making requests.

    Returns
    -------
    tuple
        Tuple containing three pandas DataFrames for each Swarm satellite (A, B, C) with the requested data.
    """
    global _token_set
    
    # Set VirES token if provided and not already set
    if token and not _token_set:
        from viresclient import set_token
        set_token(token=token, set_default=True)
        _token_set = True
    elif not token and not _token_set:
        # Try to get token from environment variable
        env_token = os.environ.get('VIRES_TOKEN')
        if env_token:
            from viresclient import set_token
            set_token(token=env_token, set_default=True)
            _token_set = True

    # Display important package versions used
    # Edit this according to what you use
    #  - this will help others to reproduce your results
    #  - it may also help to trace issues if package changes break your code
 
    # Ensure the token is set for the Vires client    
    # Create SwarmRequest objects for each satellite
    # These objects will be used to request data from the Swarm mission
    # Each request will be configured to fetch specific data products and models
    
    requestA = SwarmRequest()
    requestB = SwarmRequest()
    requestC = SwarmRequest()
    

    ### 1. Request data for Sat Alpha
    requestA.set_collection("SW_OPER_MAGA_LR_1B")
    requestA.set_products(
        measurements=[
            'F', #Magnetic intensity
            'B_NEC', #The Magnetic values are in NEC system (North, East, Centre)
            'Flags_F', #Quality Flag to validate nominal values for ASM
            'Flags_B', #Quality Flag to validate nominal values for VFM, check here for more details https://earth.esa.int/web/guest/missions/esa-eo-missions/swarm/data-handbook/level-1b-product-definitions#label-Flag-Values-of-MDR_MAG_HR
        ],

        models = [
            
            '"CHAOS_MCO_MLI_MMA" = "CHAOS-Core" + "CHAOS-Static" + "CHAOS-MMA-Primary" + "CHAOS-MMA-Secondary"'
        ],
        auxiliaries=['Kp'],
        residuals=True, #Brining the residuals.
        sampling_step="PT30S", #Get the data every 30 seconds. Trajectories often include timestamps at minute intervals
    )
   
    #2. Define an pandas dataframe to store the data request for Satellite A, based on the start Date and time.
    #You can display dsA to get an idea of how the data is requested.
    dsA = requestA.get_between(
        start_time=startDateTime,
        end_time=endDateTime, 
        show_progress = False,
        asynchronous = False
    ).as_dataframe(expand=True)
    ### End Request for Satellite Alpha
    
    ### 2. Request for Satellite Bravo, same request parameters defined by Satellite Alpha
    requestB.set_collection("SW_OPER_MAGB_LR_1B")
    requestB.set_products(
        measurements=[
            'F', #Magnetic intensity
            'B_NEC',
            'Flags_F',
            'Flags_B',

        ],
        models = [
            
            '"CHAOS_MCO_MLI_MMA" = "CHAOS-Core" + "CHAOS-Static" + "CHAOS-MMA-Primary" + "CHAOS-MMA-Secondary"'
        ],
        auxiliaries=['Kp'],
        residuals=True, 
        sampling_step="PT30S",
    )

    dsB = requestB.get_between(
        start_time= startDateTime,
        end_time= endDateTime,
        show_progress = False,
        asynchronous = False
    ).as_dataframe(expand=True)
    ### End Request for Satellite Bravo
    
    ## 3. Request for Satellite Charlie.
    requestC.set_collection("SW_OPER_MAGC_LR_1B")
    requestC.set_products(
        measurements=[
            'F',
            'B_NEC',
            'Flags_F',
            'Flags_B',

        ],
        models = [
            
            '"CHAOS_MCO_MLI_MMA" = "CHAOS-Core" + "CHAOS-Static" + "CHAOS-MMA-Primary" + "CHAOS-MMA-Secondary"'
        ],
        auxiliaries=['Kp'],
        residuals=True,
        sampling_step="PT30S", # trajectories often include timestamps at minute intervals
        #sampling_step="PT60S", #Get the data every 60 seconds.
    )

    dsC = requestC.get_between(
        start_time=startDateTime,
        end_time=endDateTime,
        show_progress = False,
        asynchronous = False
    ).as_dataframe(expand=True) 
    ### End Request for Satellite Charlie
    
    ##4. Renaming Geomagnetic components columns.
    dsA.rename(columns={"F_res_CHAOS_MCO_MLI_MMA":"F_res","B_NEC_res_CHAOS_MCO_MLI_MMA_N": "N_res", "B_NEC_res_CHAOS_MCO_MLI_MMA_E":"E_res", "B_NEC_res_CHAOS_MCO_MLI_MMA_C":"C_res"}, inplace = True)
    dsB.rename(columns={"F_res_CHAOS_MCO_MLI_MMA":"F_res","B_NEC_res_CHAOS_MCO_MLI_MMA_N": "N_res", "B_NEC_res_CHAOS_MCO_MLI_MMA_E":"E_res", "B_NEC_res_CHAOS_MCO_MLI_MMA_C":"C_res"}, inplace = True)
    dsC.rename(columns={"F_res_CHAOS_MCO_MLI_MMA":"F_res","B_NEC_res_CHAOS_MCO_MLI_MMA_N": "N_res", "B_NEC_res_CHAOS_MCO_MLI_MMA_E":"E_res", "B_NEC_res_CHAOS_MCO_MLI_MMA_C":"C_res"}, inplace = True)
    
    #5. Add the epoch column, and set that as the pandas Dataframe index. Useful to get an ID for each date and time.
    dsA['epoch'] = dsA.index
    dsA['timestamp'] = dsA.index
    dsA['epoch'] = (dsA['epoch'].astype('int64') // int(1e9)).astype(int)
    dsA.set_index("epoch", inplace=True)

    dsB['epoch'] = dsB.index
    dsB['timestamp'] = dsB.index
    dsB['epoch'] = (dsB['epoch'].astype('int64') // int(1e9)).astype(int)
    dsB.set_index("epoch", inplace=True)

    dsC['epoch'] = dsC.index
    dsC['timestamp'] = dsC.index
    dsC['epoch'] = (dsC['epoch'].astype('int64') // int(1e9)).astype(int)
    dsC.set_index("epoch", inplace=True)

    return dsA, dsB, dsC



"""
Module for fetching geomagnetic indices from HAPI servers.

This module provides functions to fetch AE (Auroral Electrojet) and SME (SuperMAG Electrojet) 
indices from their respective HAPI servers for validation and analysis of geomagnetic data.
"""

import pandas as pd
import numpy as np
from hapiclient import hapi
from hapiclient import hapitime2datetime


def get_ae_index(unique_dates, verbose=False):
    """
    Fetch AE, AL, and AU indices from NASA CDAWeb HAPI server.
    
    Parameters:
    -----------
    unique_dates : array-like
        Array of unique date strings in format 'YYYY-MM-DD'
    verbose : bool, optional
        If True, print progress information. Default is True.
    
    Returns:
    --------
    pd.DataFrame
        DataFrame with DateTime index and columns ['AE_INDEX', 'AL_INDEX', 'AU_INDEX']
        Returns None if no data was fetched successfully.
    
    Example:
    --------
    >>> dates = ['2014-08-09']
    >>> ae_data = get_ae_index(dates)
    >>> print(ae_data.head())
    """
    server = 'https://cdaweb.gsfc.nasa.gov/hapi'
    dataset = 'OMNI_HRO_1MIN'
    parameters = 'AE_INDEX,AL_INDEX,AU_INDEX'
    
    if verbose:
        print(f"Fetching AE index data for {len(unique_dates)} unique dates...")
    
    all_data = []
    for date in unique_dates:
        start = f"{date}T00:00:00Z"
        stop = f"{date}T23:59:59Z"
        
        try:
            if verbose:
                print(f"Fetching data for {date}...")
            
            d, m = hapi(server, dataset, parameters, start, stop)
            
            # HAPI returns structured data - extract fields properly
            times = hapitime2datetime(d['Time'])
            
            # Create DataFrame with proper structure
            df = pd.DataFrame({
                'DateTime': times,
                'AE_INDEX': d['AE_INDEX'],
                'AL_INDEX': d['AL_INDEX'],
                'AU_INDEX': d['AU_INDEX'],
            })
            
            df.set_index('DateTime', inplace=True)
            all_data.append(df)
            
        except Exception as e:
            if verbose:
                print(f"Error fetching data for {date}: {e}")
            continue
    
    if all_data:
        df_ae_all = pd.concat(all_data)
        if verbose:
            print(f"Successfully fetched {len(df_ae_all)} AE index records")
            print("\nSample of fetched data:")
            print(df_ae_all.head())
            
            # Show data info
            print(f"\nData covers from {df_ae_all.index.min()} to {df_ae_all.index.max()}")
            print(f"Columns available: {list(df_ae_all.columns)}")
            
            # Show some statistics
            print(f"\nAE Index statistics:")
            print(f"Mean: {df_ae_all['AE_INDEX'].mean():.1f}")
        
        return df_ae_all
    else:
        if verbose:
            print("No AE data was fetched successfully")
        return None


def get_sme_index(unique_dates, verbose=True):
    """
    Fetch SME (SuperMAG Electrojet) index from SuperMAG HAPI server.
    
    Parameters:
    -----------
    unique_dates : array-like
        Array of unique date strings in format 'YYYY-MM-DD'
    verbose : bool, optional
        If True, print progress information. Default is True.
    
    Returns:
    --------
    pd.DataFrame
        DataFrame with DateTime index and column ['SME']
        Returns None if no data was fetched successfully.
    
    Example:
    --------
    >>> dates = ['2014-08-09']
    >>> sme_data = get_sme_index(dates)
    >>> print(sme_data.head())
    """
    server_sme = 'https://supermag.jhuapl.edu/hapi'
    dataset_sme = 'indices_all'
    parameters_sme = 'SME'
    
    if verbose:
        print(f"Fetching SME index data for {len(unique_dates)} unique dates...")
    
    all_data = []
    for date in unique_dates:
        start = f"{date}T00:00:00Z"
        stop = f"{date}T23:59:59Z"
        
        try:
            if verbose:
                print(f"Fetching data for {date}...")
            
            d_2, m_2 = hapi(server_sme, dataset_sme, parameters_sme, start, stop)
            
            # HAPI returns structured data - extract fields properly
            times_2 = hapitime2datetime(d_2['Time'])
            
            # Create DataFrame with proper structure
            df = pd.DataFrame({
                'DateTime': times_2,
                'SME': d_2['SME']
            })
            
            df.set_index('DateTime', inplace=True)
            all_data.append(df)
            
        except Exception as e:
            if verbose:
                print(f"Error fetching data for {date}: {e}")
            continue
    
    if all_data:
        df_sme_all = pd.concat(all_data)
        if verbose:
            print(f"Successfully fetched {len(df_sme_all)} SME index records")
            print("\nSample of fetched data:")
            print(df_sme_all.head())
            
            # Show data info
            print(f"\nData covers from {df_sme_all.index.min()} to {df_sme_all.index.max()}")
            print(f"Columns available: {list(df_sme_all.columns)}")
            
            # Show some statistics
            print(f"\nSME Index statistics:")
            print(f"Mean: {df_sme_all['SME'].mean():.1f}")
        
        return df_sme_all
    else:
        if verbose:
            print("No SME data was fetched successfully")
        return None


def merge_indices_with_maggeo(df_csv, ae_data=None, sme_data=None, timestamp_col='timestamp'):
    """
    Merge AE and/or SME index data with MagGeo results.
    
    Parameters:
    -----------
    df_csv : pd.DataFrame
        Original MagGeo results DataFrame
    ae_data : pd.DataFrame, optional
        AE index data from get_ae_index()
    sme_data : pd.DataFrame, optional
        SME index data from get_sme_index()
    timestamp_col : str, optional
        Name of the timestamp column in df_csv. Default is 'timestamp'.
    
    Returns:
    --------
    pd.DataFrame
        Merged DataFrame with original MagGeo data and index data
    """
    # Ensure df_csv timestamp is timezone-aware to match the index data
    df_csv = df_csv.copy()  # Don't modify the original
    df_csv['timestamp_tz'] = pd.to_datetime(df_csv[timestamp_col]).dt.tz_localize('UTC')
    
    # Create a temporary dataframe for merging
    merge_df = df_csv.set_index('timestamp_tz')
    
    # Merge with AE data if provided
    if ae_data is not None:
        merge_df = merge_df.join(ae_data, how='left')
        print(f"Merged AE data - Points with AE data: {merge_df['AE_INDEX'].notna().sum()}")
    
    # Merge with SME data if provided
    if sme_data is not None:
        merge_df = merge_df.join(sme_data, how='left')
        print(f"Merged SME data - Points with SME data: {merge_df['SME'].notna().sum()}")
    
    # Clean up temporary columns
    if 'date' in merge_df.columns:
        merge_df.drop(columns=['date'], inplace=True)
    
    return merge_df



from typing import Any, Dict
import pandas as pd
from .auxiliaryfunctions import distance_to_GPS, Kradius, DistJ, DfTime_func

def st_idw_process(
    gps_lat: float,
    gps_long: float,
    gps_altitude: float,
    gps_datetime: pd.Timestamp,
    gps_epoch: int,
    swarm_a: pd.DataFrame,
    swarm_b: pd.DataFrame,
    swarm_c: pd.DataFrame,
    dt_seconds: int = 14400
) -> Dict[str, Any]:
    """
    Interpolate Swarm residuals to a GPS point using spatiotemporal IDW.
    Returns a dictionary with interpolated values and metadata.
    """
    # This function performs the spatiotemporal inverse distance weighting (IDW) interpolation
    # for the Swarm residuals based on the GPS trajectory.
    # Interpolation of the Swarm Residuals., NEC interpolated residuals for each GPS Point. Quality flags filters.
    # Input:  GPS Trajectory columns, SwarmDataDF
    # Output: GPS Trajectory + ResidualsInterpolated

    # [REMOVE before Package Release]
    #print(f"=====")
    #print(f"[DEBUG] ST IDW Interpolation process with params: {gps_lat},{gps_long},{gps_altitude}, {gps_datetime}, {gps_epoch}, {dt_seconds}, {swarm_a.shape},{swarm_b.shape},{swarm_c.shape}")
    #print(f"=====")
    
    # 1. Running the DfTime_func function to filter by the defined DeltaTime.
    tk_a = DfTime_func(swarm_a, gps_epoch, dt_seconds)

    # [REMOVE before Package Release]
    #print(f"[DEBUG] tk_a len after DfTime_func: {len(tk_a)}")
    #print(f"=====")
    #print(f"=====")

    tk_b = DfTime_func(swarm_b, gps_epoch, dt_seconds)
    tk_c = DfTime_func(swarm_c, gps_epoch, dt_seconds)
    
    # [REMOVE before Package Release]
    #print(f"[DEBUG] tk_a shape after time filter: {tk_a.shape}")
    #print(tk_a.head())
    #tk_a.to_csv(f"temp_data/tk_a_{gps_epoch}.csv")
    #print(f"[DEBUG] tk_b shape after time filter: {tk_b.shape}")
    #print(tk_b.head())
    #tk_b.to_csv(f"temp_data/tk_b_{gps_epoch}.csv")
    #print(f"[DEBUG] tk_c shape after time filter: {tk_c.shape}")
    #print(tk_c.head())
    #tk_c.to_csv(f"temp_data/tk_c_{gps_epoch}.csv")

    #2. Computing the dt as the difference between the datetime and the datetime from swarm point. At this point
    # We have filtered the swarm point by time.

    # [REMOVE before Package Release]
    #print(f"=====")
    #print(f"[DEBUG] Computing time difference for tk_a, tk_b, tk_c with gps_epoch: {gps_epoch} and dt_seconds: {dt_seconds}")
    #print(f"=====")

    tk_a = tk_a.copy()
    tk_a['dT'] = (gps_epoch - tk_a.index)
    tk_b = tk_b.copy()
    tk_b['dT'] = (gps_epoch - tk_b.index) 
    tk_c = tk_c.copy()
    tk_c['dT'] = (gps_epoch - tk_c.index)

    # [REMOVE before Package Release]
    #print(f"[DEBUG] tk_a schema and shape: {tk_a.shape}")
    #print(tk_a.head(5))
    #tk_a.to_csv(f"temp_data/tk_a_{gps_epoch}_dT.csv")
    #tk_b.to_csv(f"temp_data/tk_b_{gps_epoch}_dT.csv")
    #tk_c.to_csv(f"temp_data/tk_c_{gps_epoch}_dT.csv")
    #print(f"=====")


    #3. Compute spatial distance (ds)
    ### Parsing the requires parameters for distance_to_GPS function

    # [REMOVE before Package Release]
    #print(f"=====")
    #print(f"[DEBUG] Computing spatial distance for tk_a, tk_b, tk_c with gps_lat: {gps_lat}, gps_long: {gps_long}, tk_a['Latitude'], tk_a['Longitude']")
    #print(f"=====")

    tk_a['distance'] = distance_to_GPS(gps_lat, gps_long, tk_a['Latitude'], tk_a['Longitude'])
    tk_b['distance'] = distance_to_GPS(gps_lat, gps_long, tk_b['Latitude'], tk_b['Longitude'])
    tk_c['distance'] = distance_to_GPS(gps_lat, gps_long, tk_c['Latitude'], tk_c['Longitude'])
    
    # [REMOVE before Package Release]
    #print(f"[DEBUG] tk_a schema and shape: {tk_a.shape}")
    #print(tk_a.head(5))
    #tk_a.to_csv(f"temp_data/tk_a_{gps_epoch}_distance.csv")

    #4. Compute radius (r)
    # [REMOVE before Package Release]
    #print(f"=====")
    #print(f"[DEBUG] Computing radius for gps_lat: {gps_lat}")
    #print(f"=====")
    
    tk_a['r'] = Kradius(gps_lat)
    # [REMOVE before Package Release]
    #print(f"[DEBUG] tk_a after Computing radius schema and shape: {tk_a.shape}")
    #print(tk_a.head(5))
    #tk_a.to_csv(f"temp_data/tk_a_{gps_epoch}_radius.csv")
    tk_b['r'] = Kradius(gps_lat)
    tk_c['r'] = Kradius(gps_lat)
    
   
    # [REMOVE before Package Release]
    #print(f"=====")
    #print(f"[DEBUG] Filtering rows that only fall into the computed R value")
    #print(f"=====")
    
    #5. Filtering rows that only fall into the computed R value.
    k_a = tk_a[tk_a['distance'] <= tk_a['r']]
    k_b = tk_b[tk_b['distance'] <= tk_b['r']]
    k_c = tk_c[tk_c['distance'] <= tk_c['r']]
    # [REMOVE before Package Release]
    #print(f"[DEBUG] filtered data  after space filter: {k_a.shape}")
    #k_a.to_csv(f"temp_data/k_a_{gps_epoch}.csv")
    #print(f"[DEBUG] k_b shape after space filter: {k_b.shape}")
    #k_b.to_csv(f"temp_data/k_b_{gps_epoch}.csv")
    #print(f"[DEBUG] k_c shape after space filter: {k_c.shape}")
    #k_c.to_csv(f"temp_data/k_c_{gps_epoch}.csv")
    #print(f"[DEBUG] tk_a distances: {tk_a['distance'].describe()}")
    #print(f"[DEBUG] tk_a radius: {tk_a['r'].iloc[0] if not tk_a.empty else 'N/A'}")

    #6. Quality filtering, Filtering Bad Points, using quality flags
    def filter_flags(df):
        df = df[df['F_res'].between(-2000, 2000)]
        df = df[df['Flags_F'].between(0, 1)]
        df = df[df['Flags_B'].between(0, 1)]
        return df
    
    # [REMOVE before Package Release]
    #print(f"[DEBUG] k_a before quality filter: {k_a.shape}")
    #print(f"[DEBUG] k_b before quality filter: {k_b.shape}")
    #print(f"[DEBUG] k_c before quality filter: {k_c.shape}")
    
    k_a = filter_flags(k_a)
    k_b = filter_flags(k_b)
    k_c = filter_flags(k_c)

    # [REMOVE before Package Release]
    #print(f"[DEBUG] k_a shape after quality filter: {k_a.shape}")
    #k_a.to_csv(f"temp_data/k_a_quality_{gps_epoch}.csv")
    #print(f"[DEBUG] k_b shape after quality filter: {k_b.shape}")
    #print(k_b.head())
    #k_b.to_csv(f"temp_data/k_b_quality_{gps_epoch}.csv")
    #print(f"[DEBUG] k_c shape after quality filter: {k_c.shape}")
    #print(k_c.head())
    #k_c.to_csv(f"temp_data/k_c_quality_{gps_epoch}.csv")

    #7.Combining the three satellited measures into a bigger dataframe that store all the Swarm points that were filtered. 
    swarm_filtered = pd.concat([k_a, k_b, k_c], keys=['A', 'B', 'C'], sort=False)
    
    # [REMOVE before Package Release]
    #print(f"[DEBUG] swarm_filtered_combined_ABC shape: {swarm_filtered.shape}")
    #print(swarm_filtered.head(5))
    
    #8.  Computing the minimum and average distance and the Kp index average.
    min_dist = swarm_filtered['distance'].min()
    avg_dist = swarm_filtered['distance'].mean()
    kp_avg = swarm_filtered['Kp'].mean()

    # [REMOVE before Package Release]
    #swarm_filtered.to_csv(f"temp_data/swarm_filtered_{gps_epoch}.csv")
    #print(f"=====")
    #print(f"[DEBUG] swarm_filtered shape: {swarm_filtered.shape}")
    #print(swarm_filtered.head(5))
    #print(f"=====")

    #8. Compute weights - Computing the d (hypotenuse computed from the edges ds, dt values
    ds = swarm_filtered['distance']
    r = swarm_filtered['r']
    dt = swarm_filtered['dT']
    swarm_filtered['Dj'] = DistJ(ds, r, dt, dt_seconds)
    swarm_filtered['W'] = 1 / (swarm_filtered['Dj'] ** 2)
    
    # [REMOVE before Package Release]
    #print(f"[DEBUG] swarm_filtered after Dj and W computation values: {swarm_filtered.shape}")
    #print(swarm_filtered.head(5))
    #swarm_filtered.to_csv(f"temp_data/swarm_filtered_{gps_epoch}_weights.csv")
    #print(f"=====")

    #9. Computing the Sum of weights    
    sum_w = swarm_filtered['W'].sum()
    #10. Distribution of weights
    swarm_filtered['Wj'] = swarm_filtered['W'] / sum_w

    # [REMOVE before Package Release]
    #print(f"=====")
    #print(f"[DEBUG] Weights sum: {sum_w}, Points: {len(swarm_filtered)}")
    #print(swarm_filtered[['distance', 'r', 'dT', 'Dj', 'W', 'Wj', 'N_res', 'E_res', 'C_res']].head())
    #print(f"=====")

    #9. Weighted interpolation - Computing the Magnetic component based on the weights previous weights. 
    n_res = (swarm_filtered['Wj'] * swarm_filtered['N_res']).sum()
    e_res = (swarm_filtered['Wj'] * swarm_filtered['E_res']).sum()
    c_res = (swarm_filtered['Wj'] * swarm_filtered['C_res']).sum()
    
    # [REMOVE before Package Release]
    #print(f"=====")
    #print(f"[DEBUG] Final part Interpolated N_res: {n_res}, E_res: {e_res}, C_res: {c_res}")
    #print(f"=====")
    
    #15. Write the results into an array that will be a dictionary for the final dataframe.
    return {
        'Latitude': gps_lat,
        'Longitude': gps_long,
        'Altitude': gps_altitude,
        'DateTime': gps_datetime,
        'N_res': n_res,
        'E_res': e_res,
        'C_res': c_res,
        'TotalPoints': len(swarm_filtered),  #Calculating the number of points per satellite that have passed the Space and Time Kernels.
        'Minimum_Distance': min_dist,
        'Average_Distance': avg_dist,
        'Kp': kp_avg
    }




"""
Configuration-based debug system for MagGeo.
"""
import yaml
from dataclasses import dataclass
from typing import Optional
from pathlib import Path

@dataclass
class DebugConfig:
    """Debug configuration settings."""
    enabled: bool = False
    log_level: str = "INFO"  # DEBUG, INFO, WARNING, ERROR
    save_temp_files: bool = False
    temp_dir: str = "temp_data"
    log_file: str = "maggeo.log"
    
    # Specific debug options
    log_parameters: bool = True
    log_gps_data: bool = True
    log_swarm_data: bool = True
    log_interpolation: bool = True
    log_chaos_values: bool = True
    
    # File saving options
    save_gps_data: bool = True
    save_swarm_data: bool = True
    save_interpolation: bool = True
    save_chaos_results: bool = True

def load_debug_config(config_path: Optional[str] = None) -> DebugConfig:
    """Load debug configuration from YAML file or use defaults."""
    if config_path and Path(config_path).exists():
        with open(config_path, 'r') as f:
            config_dict = yaml.safe_load(f).get('debug', {})
        return DebugConfig(**config_dict)
    return DebugConfig()

def create_debug_config_template(output_path: str = "debug_config.yml"):
    """Create a template debug configuration file."""
    template = {
        'debug': {
            'enabled': False,
            'log_level': 'INFO',
            'save_temp_files': False,
            'temp_dir': 'temp_data',
            'log_file': 'maggeo.log',
            'log_parameters': True,
            'log_gps_data': True,
            'log_swarm_data': True,
            'log_interpolation': True,
            'log_chaos_values': True,
            'save_gps_data': True,
            'save_swarm_data': True,
            'save_interpolation': True,
            'save_chaos_results': True
        }
    }
    
    with open(output_path, 'w') as f:
        yaml.dump(template, f, default_flow_style=False, indent=2)
    
    print(f"Debug configuration template created at: {output_path}")
    print("Edit this file and set 'enabled: true' to activate debug mode.")

if __name__ == "__main__":
    create_debug_config_template()



# maggeo/__init__.py
"""
MagGeo: Data fusion library for annotating GPS trajectories with geomagnetic satellite data from Swarm mission from ESA.
Authors: Fernando Benitez-Paez, Urška Demšar, Jed Long, Ciaran Beggan

This package materialise a data fusion method to annotate GPS trajectories with geomagnetic data from the Swarm satellite mission.
It includes functions for data retrieval, processing, and annotation, as well as debugging utilities.
For more information, please refer to the paper:

Benitez-Paez, F., Brum-Bastos, V.d., Beggan, C.D. et al. 
Fusion of wildlife tracking and satellite geomagnetic data for the study of animal migration. 
Mov Ecol 9, 31 (2021). https://doi.org/10.1186/s40462-021-00268-4

"""

__version__ = "0.2.1"
__author__ = "Fernando Benitez-Paez, Urška Demšar, Jed Long, Ciaran Beggan"
__email__ = "Fernando.Benitez@st-andrews.ac.uk"

# Lazy imports to avoid immediate dependency loading
def get_main_function():
    """Get the main annotation function with lazy import."""
    try:
        from .core import annotate_gps_with_geomag
        return annotate_gps_with_geomag
    except ImportError as e:
        raise ImportError(f"Could not import core functionality. Please ensure all dependencies are installed: {e}")

# Only expose main function, date utilities, indices, parallel functions, and swarm data manager to avoid import issues
__all__ = ['annotate_gps_with_geomag', 'identify_unique_dates', 'get_ae_index', 'get_sme_index', 'merge_indices_with_maggeo', 
           'SwarmDataManager', 'download_swarm_data_for_trajectory', 'load_swarm_data', 
           'parallel_row_handler', 'parallel_st_idw_process', 'parallel_chaos_ground_values', '__version__']

# Make main function, date utilities, indices, parallel functions, and swarm data manager available
try:
    from .core import annotate_gps_with_geomag
    from .date_utils import identify_unique_dates
    from .parallel_processing import parallel_row_handler
    from .interpolation import parallel_st_idw_process
    from .chaos import parallel_chaos_ground_values
    from .indices import get_ae_index, get_sme_index, merge_indices_with_maggeo
    from .swarm_data_manager import SwarmDataManager, download_swarm_data_for_trajectory, load_swarm_data
except ImportError:
    # Provide informative error message
    def annotate_gps_with_geomag(*args, **kwargs):
        raise ImportError(
            "MagGeo core functionality is not available. "
            "Please ensure all dependencies are installed:\n"
            "pip install viresclient chaosmagpy pandas netCDF jupyterlab \n"
            "or install MagGeo with conda using the provided environment.yml"
        )
    
    def identify_unique_dates(*args, **kwargs):
        raise ImportError(
            "MagGeo core functionality is not available. "
            "Please ensure all dependencies are installed:\n"
            "pip install viresclient chaosmagpy pandas netCDF jupyterlab hapiclient \n"
            "or install MagGeo with conda using the provided environment.yml"
        )
    
    def get_ae_index(*args, **kwargs):
        raise ImportError(
            "MagGeo indices functionality is not available. "
            "Please ensure all dependencies are installed:\n"
            "pip install hapiclient pandas numpy \n"
            "or install MagGeo with conda using the provided environment.yml"
        )
    
    def get_sme_index(*args, **kwargs):
        raise ImportError(
            "MagGeo indices functionality is not available. "
            "Please ensure all dependencies are installed:\n"
            "pip install hapiclient pandas numpy \n"
            "or install MagGeo with conda using the provided environment.yml"
        )
    
    def merge_indices_with_maggeo(*args, **kwargs):
        raise ImportError(
            "MagGeo indices functionality is not available. "
            "Please ensure all dependencies are installed:\n"
            "pip install hapiclient pandas numpy \n"
            "or install MagGeo with conda using the provided environment.yml"
        )


"""
Date utilities for MagGeo trajectory processing.

This module contains functions for processing and analyzing dates in GPS trajectories,
particularly for determining the unique dates needed for Swarm satellite data download.

Main function:
- identify_unique_dates(): Analyzes GPS trajectories to determine unique dates needed for Swarm data download. 
Swarm satellite mission and its data are polar orbits and typically available in 1-day intervals (every second),
so this function identifies a buffer dates for early morning and late evening GPS points to ensure complete coverage that matches
the time window of 4 hours before and after the GPS point.

See the paper:
- Benitez, F., et al. (2021). Fusion of wildlife tracking and satellite geomagnetic data for the study of animal migration. 
https://doi.org/10.1186/s40462-021-00268-4
  
"""

from typing import Dict, Any
import pandas as pd
from .debug import get_debugger


def identify_unique_dates(gps_df: pd.DataFrame) -> pd.DataFrame:
    """
    Identify unique dates for Swarm data download process.
    
    This function analyzes the GPS trajectory to determine all unique dates
    that need Swarm data. It includes buffer dates for early morning and
    late evening GPS points to ensure complete coverage.
    
    Args:
        gps_df (pd.DataFrame): GPS DataFrame with 'dates' and 'times' columns
        
    Returns:
        pd.DataFrame: DataFrame with unique dates and metadata for Swarm data download
                     Columns: ['date', 'is_buffer_date', 'buffer_type', 'original_date']
    """
    debugger = get_debugger()
    debugger.log("Starting unique dates identification process")
    
    # Create time string column for filtering
    gps_df['time_str'] = gps_df['times'].apply(lambda t: t.strftime('%H:%M:%S'))
    
    # Identify early morning and late evening points
    early_mask = gps_df['time_str'] < '04:00:00'
    late_mask = gps_df['time_str'] > '20:00:00'
    
    debugger.log(f"Found {early_mask.sum()} early morning points (before 04:00:00)")
    debugger.log(f"Found {late_mask.sum()} late evening points (after 20:00:00)")
    
    # Collect all dates needed
    dates_data = []
    
    # Add original dates
    for date in gps_df['dates'].unique():
        dates_data.append({
            'date': date,
            'is_buffer_date': False,
            'buffer_type': None,
            'original_date': date
        })
    
    # Add buffer dates for early morning points (previous day)
    for date in gps_df.loc[early_mask, 'dates'].unique():
        buffer_date = date - pd.Timedelta(days=1)
        dates_data.append({
            'date': buffer_date,
            'is_buffer_date': True,
            'buffer_type': 'early_morning',
            'original_date': date
        })
    
    # Add buffer dates for late evening points (next day)
    for date in gps_df.loc[late_mask, 'dates'].unique():
        buffer_date = date + pd.Timedelta(days=1)
        dates_data.append({
            'date': buffer_date,
            'is_buffer_date': True,
            'buffer_type': 'late_evening',
            'original_date': date
        })
    
    # Create DataFrame and remove duplicates
    unique_dates_df = pd.DataFrame(dates_data)
    unique_dates_df = unique_dates_df.drop_duplicates(subset=['date']).sort_values('date').reset_index(drop=True)
    
    debugger.log(f"Total unique dates identified: {len(unique_dates_df)}")
    debugger.log(f"Original trajectory dates: {len(gps_df['dates'].unique())}")
    debugger.log(f"Buffer dates added: {unique_dates_df['is_buffer_date'].sum()}")
    debugger.log_unique_dates(unique_dates_df['date'])
    
    # Clean up temporary column
    gps_df.drop(columns=['time_str'], inplace=True)
    
    return unique_dates_df



# 0. Get the GPS track in a CSV format.
# Input: csv file store in the data folder, validate if there is a altitude attribute.
# Output: GPS Data as pandas DF.
import os
import pandas as pd

def get_gps_data(
        data_dir,
        gpsfilename,
        lat_col,
        lon_col,
        datetime_col,
        altitude_col='',
        return_original_cols=False
    ) -> pd.DataFrame:
    """Get GPS data from a CSV file and return it as a pandas DataFrame.
    This function reads a CSV file containing GPS data, renames the columns to standard names,
    and ensures that the altitude column is present. If the altitude column is not provided,
    it defaults to an empty string and sets the altitude to 0 for all entries.
    If the altitude is not provided, it will be set to 0 for all entries.

    Parameters
    ----------
    data_dir : str
        Directory where the GPS data CSV file is located.
    gpsfilename : str
        Name of the GPS data CSV file.
    lat_col : str
        Name of the column containing latitude data.
    lon_col : str
        Name of the column containing longitude data.
    datetime_col : str
        Name of the column containing date and time data.
    altitude_col : str
        Name of the column containing altitude data. If not provided, defaults to an empty string.
        If altitude is not provided, it will be set to 0 for all entries
    return_original_cols : bool, optional
        If True, also return the original DataFrame as read from CSV (default False).

    Returns
    -------
    pd.DataFrame
        DataFrame containing the GPS data with columns renamed to 'gpsLat', 'gpsLong',
        'gpsDateTime', and 'gpsAltitude'. If altitude is not provided, 'gpsAltitude' will be set to 0 for all entries.
    """
    # Read the original CSV file without any modifications
    original_nfp = pd.read_csv(os.path.join(data_dir, gpsfilename), encoding='utf-8', dayfirst=True)

    # Check if altitude is provided, if not, set it to an empty string
    if not altitude_col:
        nfp = pd.read_csv(
            os.path.join(data_dir, gpsfilename),
            parse_dates=[0], encoding='utf-8', dayfirst=True,
            usecols=[lat_col, lon_col, datetime_col]
        )
        nfp['gpsAltitude'] = 0
        rename_dict = {lat_col: 'gpsLat', lon_col: 'gpsLong', datetime_col: 'gpsDateTime'}
    else:
        # If altitude is provided, read it from the CSV file (include altitude_col in usecols)
        nfp = pd.read_csv(
            os.path.join(data_dir, gpsfilename),
            parse_dates=[0], encoding='utf-8', dayfirst=True,
            usecols=[lat_col, lon_col, datetime_col, altitude_col]
        )
        nfp.loc[(nfp[altitude_col] < 0) | (nfp[altitude_col].isnull()), altitude_col] = 0
        rename_dict = {lat_col: 'gpsLat', lon_col: 'gpsLong', datetime_col: 'gpsDateTime', altitude_col: 'gpsAltitude'}

    nfp.rename(columns=rename_dict, inplace=True)
    # Convert the gpsDateTime to datetime python object
    nfp['gpsDateTime'] = pd.to_datetime(nfp['gpsDateTime'])
    nfp['gpsDateTime'] = nfp['gpsDateTime'].map(lambda x: x.replace(second=0))
    nfp['gpsLat'] = nfp['gpsLat'].astype(float)
    nfp['gpsLong'] = nfp['gpsLong'].astype(float)
    # Adding new column epoch, will be useful to compare the date&time o each gps point against the gathered Swarm data points
    #nfp['epoch'] = nfp['gpsDateTime'].view('int64') // 1_000_000_000
    nfp['epoch'] = nfp['gpsDateTime'].astype('int64') // 1e9
    nfp['epoch'] = nfp['epoch'].astype(int)
    # Computing Date and Time columns
    nfp['dates'] = nfp['gpsDateTime'].dt.date
    nfp['times'] = nfp['gpsDateTime'].dt.time

    if return_original_cols:
        # Return the original DataFrame as well
        return original_nfp.copy()

    return nfp



from typing import Dict, Any
import pandas as pd
import numpy as np
from .gps import get_gps_data
from .swarm import get_swarm_residuals
from .swarm_data_manager import SwarmDataManager
from .interpolation import st_idw_process
from .chaos import chaos_ground_values
from .parallel_processing import parallel_maggeo_annotation
from .debug import get_debugger
from .date_utils import identify_unique_dates
import datetime as dt
from tqdm import tqdm

def annotate_gps_with_geomag(
        params: Dict[str, Any], 
        use_swarm_manager: bool = False, 
        use_parallel: bool = False, 
        n_cores: int = None
        ) -> pd.DataFrame:
    """
    MagGeo annotation Pipeline.
    
    Parameters
    ----------
    params : dict 
        Dictionary with keys for GPS file, columns, and token.
    use_swarm_manager : bool, default False
        Whether to use the SwarmDataManager for more efficient data handling.
        If True, data will be saved locally and can be reused later.
    use_parallel : bool, default False
        Whether to use parallel processing for interpolation and CHAOS calculations.
        Significantly speeds up processing for large datasets.
    n_cores : int, optional
        Number of cores to use for parallel processing. If None, uses all available cores.
        
    Returns
    -------
    pd.DataFrame
        Annotated DataFrame with geomagnetic data.
    """

    # Initialize debugger
    debugger = get_debugger()
    debugger.log_parameters(params)

    # 0. Get the GPS track in a CSV format.
    # Input: csv file store in the data folder, validate if there is a altitude attribute.
    # Output: GPS Data as pandas DF.

    gps_df = get_gps_data(
        data_dir=params['data_dir'],
        gpsfilename=params['gpsfilename'],
        lat_col=params['lat_col'],
        lon_col=params['long_col'],
        datetime_col=params['datetime_col'],
        altitude_col=params.get('altitude_col', None)
    )
    
    debugger.print_dataframe_info(gps_df, "GPS")
    debugger.save_dataframe(gps_df, "gps_loaded.csv", "Loaded GPS data")

    # 2. Identify unique dates for Swarm data download process (including buffer dates)
    unique_dates_df = identify_unique_dates(gps_df) 
    unique_dates = unique_dates_df['date']  # Extract dates

    # 3. Download Swarm data for each date
    # For each day in the trajectory, Get the Swarm Data and Residuals: Get_Swarm_and_residuals
    # Input:  Date and Time variables
    # Output: Swarm DF for each Sat, including the residuals, and Quality Flags.
    
    if use_swarm_manager:
        # Use the SwarmDataManager for more efficient data handling
        debugger.log("Using SwarmDataManager for Swarm data download")
        swarm_manager = SwarmDataManager(
            data_dir=params.get('swarm_data_dir', 'temp_data/swarm_data'),
            file_format=params.get('swarm_file_format', 'parquet'),
            token=params.get('token')
        )
        swarm_a, swarm_b, swarm_c = swarm_manager.download_for_trajectory(
            gps_df, 
            resume=params.get('resume_swarm_download', True)
        )
    else:
        # Use the original method released in the first version of MagGeo. Ill keep it like this, so in case I need to change the logic, 
        # I can do it without affecting the SwarmDataManager.
        # But if works I will remove it in the next version, and process everything through the SwarmDataManager.

        swarm_a_list, swarm_b_list, swarm_c_list = [], [], []
        hours_added = dt.timedelta(hours = 23, minutes=59, seconds=59)  # Add 23 hours and 59 minutes to cover the full day of data

        for d in tqdm (unique_dates, desc="Getting Swarm Data"):

            debugger.log(f"Requesting data for date: {d}")
            startdate = dt.datetime.combine(d, dt.datetime.min.time())
            enddate = startdate + hours_added
            debugger.log_date_range(d, startdate, enddate)
            
            # Get Swarm data for the specified date range
            a, b, c = get_swarm_residuals(startdate, enddate, params.get('token'))
            
            debugger.log_swarm_data("A", d, a)
            debugger.log_swarm_data("B", d, b) 
            debugger.log_swarm_data("C", d, c)
            
            swarm_a_list.append(a)
            swarm_b_list.append(b)
            swarm_c_list.append(c)
            
        swarm_a = pd.concat(swarm_a_list)
        debugger.log_swarm_concat("A", swarm_a)
        
        swarm_b = pd.concat(swarm_b_list)
        debugger.log_swarm_concat("B", swarm_b)
        
        swarm_c = pd.concat(swarm_c_list)
        debugger.log_swarm_concat("C", swarm_c)


    # 4. Annotate with GeoMagnetic components each GPS point
    if use_parallel:
        # Use parallel processing for faster computation
        debugger.log("Using parallel processing")
        
        # Use the corrected parallel pipeline that follows MagGeo logic
        annotated_df = parallel_maggeo_annotation(
            gps_df, swarm_a, swarm_b, swarm_c,
            dt_seconds=params.get('dt_seconds', 14400),
            n_cores=n_cores,
            chunk_size=params.get('chunk_size', None)
        )
        
        debugger.log_interpolation_result(annotated_df)
        debugger.log_chaos_result(annotated_df)
        
    else:
        # This is the original sequential processing logic, The idea behind is to process each GPS point with the complete Swarm data
        # based on a sequential IDW interpolation process., for short routes it migth me more easy to understand and debug.
        # it also do not save any Swarm data, so does not require the SwarmDataManager, but is not as efficient as the parallel processing
        
        results = []
        for _, row in tqdm (gps_df.iterrows(), total=gps_df.shape[0], desc="Annotating your GPS Trajectory"):
            try:
                result = st_idw_process(
                    row['gpsLat'], row['gpsLong'], row['gpsAltitude'],
                    row['gpsDateTime'], row['epoch'],
                    swarm_a, swarm_b, swarm_c
                )
            except:
                print("Ups!.That was a bad Swarm Point, let's keep working with the next point")
                result = {
                    'Latitude': row['gpsLat'],
                    'Longitude': row['gpsLong'],
                    'Altitude': row['gpsAltitude'],
                    'DateTime': row['gpsDateTime'],
                    'N_res': float('nan'),
                    'E_res': float('nan'),
                    'C_res': float('nan'),
                    'TotalPoints': 0,
                    'Minimum_Distance': float('nan'),
                    'Average_Distance': float('nan'),
                    'Kp': float('nan')
                }
            results.append(result)
        annotated_df = pd.DataFrame(results)
        debugger.log_interpolation_result(annotated_df)

        # 5. Compute CHAOS ground values
        debugger.log("Computing CHAOS ground values")

        X_obs, Y_obs, Z_obs, X_obs_internal, Y_obs_internal, Z_obs_internal = chaos_ground_values(annotated_df)
        annotated_df['N'] = X_obs
        debugger.log_chaos_values("N (X_obs)", X_obs)
        
        annotated_df['E'] = Y_obs
        debugger.log_chaos_values("E (Y_obs)", Y_obs)
        
        annotated_df['C'] = Z_obs
        debugger.log_chaos_values("C (Z_obs)", Z_obs)
        
        annotated_df['N_Obs'] = X_obs_internal
        debugger.log_chaos_values("N_Obs (X_obs_internal)", X_obs_internal)
        
        annotated_df['E_Obs'] = Y_obs_internal
        debugger.log_chaos_values("E_Obs (Y_obs_internal)", Y_obs_internal)
        
        annotated_df['C_Obs'] = Z_obs_internal
        debugger.log_chaos_values("C_Obs (Z_obs_internal)", Z_obs_internal)
        
        # 5b. Compute additional geomagnetic components like H, Declination, the intensity and inclination
        annotated_df['H'] = np.sqrt((annotated_df['N']**2 + annotated_df['E']**2))
        annotated_df['D'] = np.degrees(np.arctan2(annotated_df['E'], annotated_df['N']))
        annotated_df['I'] = np.degrees(np.arctan2(annotated_df['C'], annotated_df['H']))
        annotated_df['F'] = np.sqrt((annotated_df['N']**2 + annotated_df['E']**2 + annotated_df['C']**2))
        
        debugger.log_chaos_result(annotated_df)
    
    #Drop duplicated and unnecessary columns. 
    columns_to_drop = ['Latitude', 'Longitude', 'DateTime', 'Altitude', 'N_res', 'E_res', 'C_res']
    annotated_df.drop(columns=[col for col in columns_to_drop if col in annotated_df.columns], inplace=True)
    
    debugger.log("Merging GPS DataFrame with annotated DataFrame")

    # 6. Merge GPS and annotated DataFrames
    gps_df_original = get_gps_data (data_dir=params['data_dir'],
                                    gpsfilename=params['gpsfilename'],
                                    lat_col=params['lat_col'],
                                    lon_col=params['long_col'],
                                    datetime_col=params['datetime_col'],
                                    altitude_col=params.get('altitude_col', None),
                                    return_original_cols=True)
    
    debugger.print_dataframe_info(gps_df_original, "Original GPS")
    debugger.print_dataframe_info(annotated_df, "Annotated")

    # Merge the two DataFrames
    MagGeoResult = pd.concat([gps_df_original, annotated_df], axis=1)
    debugger.print_dataframe_info(MagGeoResult, "MagGeoResult")
    debugger.save_maggeo_results(MagGeoResult, "MagGeoResult.csv", "MagGeo Results")

    print("Congrats! MagGeo has processed your GPS trajectory.")
    return MagGeoResult



"""
Command Line Interface for MagGeo package.

This module provides CLI commands for GPS trajectory annotation with geomagnetic data.
"""

import click
import pandas as pd
import os
from pathlib import Path
from .core import annotate_gps_with_geomag
from .swarm_data_manager import SwarmDataManager
from .parallel_processing import parallel_maggeo_annotation
from . import __version__


@click.group()
@click.version_option(version=__version__, prog_name="maggeo")
def main():
    """MagGeo: GPS trajectory annotation with geomagnetic data from Swarm satellites."""
    pass


@main.command()
@click.argument('input_file', type=click.Path(exists=True))
@click.option('--output', '-o', type=click.Path(), 
              help='Output file path (default: input_file_annotated.csv)')
@click.option('--lat-col', default='latitude', 
              help='Column name for latitude (default: latitude)')
@click.option('--lon-col', default='longitude', 
              help='Column name for longitude (default: longitude)')
@click.option('--datetime-col', default='datetime', 
              help='Column name for datetime (default: datetime)')
@click.option('--altitude-col', default=None, 
              help='Column name for altitude (optional)')
@click.option('--parallel', is_flag=True, 
              help='Use parallel processing for large files')
@click.option('--chunk-size', default=1000, type=int,
              help='Chunk size for parallel processing (default: 1000)')
def annotate(input_file, output, lat_col, lon_col, datetime_col, altitude_col, parallel, chunk_size):
    """Annotate GPS trajectory with geomagnetic data."""
    
    if output is None:
        input_path = Path(input_file)
        output = input_path.parent / f"{input_path.stem}_annotated{input_path.suffix}"
    
    click.echo(f"Reading GPS data from: {input_file}")
    
    try:
        # Read GPS data
        gps_df = pd.read_csv(input_file)
        
        # Validate required columns
        required_cols = [lat_col, lon_col, datetime_col]
        missing_cols = [col for col in required_cols if col not in gps_df.columns]
        if missing_cols:
            raise click.ClickException(f"Missing required columns: {missing_cols}")
        
        click.echo(f"Processing {len(gps_df)} GPS points...")
        
        if parallel and len(gps_df) > chunk_size:
            click.echo("Using parallel processing...")
            annotated_df = parallel_maggeo_annotation(
                gps_df, 
                lat_col=lat_col,
                lon_col=lon_col,
                datetime_col=datetime_col,
                altitude_col=altitude_col,
                chunk_size=chunk_size
            )
        else:
            # Use single-threaded processing
            annotated_df = annotate_gps_with_geomag(
                gps_df,
                lat_col=lat_col,
                lon_col=lon_col,
                datetime_col=datetime_col,
                altitude_col=altitude_col
            )
        
        # Save results
        annotated_df.to_csv(output, index=False)
        click.echo(f"Annotated trajectory saved to: {output}")
        click.echo(f"Added {len(annotated_df.columns) - len(gps_df.columns)} geomagnetic columns")
        
    except Exception as e:
        raise click.ClickException(f"Error processing file: {str(e)}")


@main.command()
@click.option('--start-date', required=True, type=str,
              help='Start date (YYYY-MM-DD format)')
@click.option('--end-date', required=True, type=str,
              help='End date (YYYY-MM-DD format)')
@click.option('--output-dir', '-o', type=click.Path(),
              help='Output directory for Swarm data (default: ./swarm_data)')
@click.option('--satellites', default='A,B,C',
              help='Comma-separated list of Swarm satellites (default: A,B,C)')
@click.option('--data-type', default='MAG',
              help='Swarm data type (default: MAG)')
def swarm(start_date, end_date, output_dir, satellites, data_type):
    """Download and manage Swarm satellite data."""
    
    if output_dir is None:
        output_dir = './swarm_data'
    
    satellite_list = [sat.strip() for sat in satellites.split(',')]
    
    click.echo(f"Downloading Swarm data from {start_date} to {end_date}")
    click.echo(f"Satellites: {satellite_list}")
    click.echo(f"Output directory: {output_dir}")
    
    try:
        manager = SwarmDataManager()
        
        for satellite in satellite_list:
            click.echo(f"Processing Swarm-{satellite}...")
            data = manager.download_swarm_data(
                start_date=start_date,
                end_date=end_date,
                satellite=satellite,
                data_type=data_type
            )
            
            output_file = os.path.join(output_dir, f"swarm_{satellite}_{start_date}_{end_date}.parquet")
            os.makedirs(output_dir, exist_ok=True)
            data.to_parquet(output_file)
            
            click.echo(f"Saved {len(data)} records to {output_file}")
        
        click.echo("Swarm data download completed!")
        
    except Exception as e:
        raise click.ClickException(f"Error downloading Swarm data: {str(e)}")


@main.command()
@click.argument('input_file', type=click.Path(exists=True))
def validate(input_file):
    """Validate GPS trajectory file format."""
    
    click.echo(f"Validating GPS file: {input_file}")
    
    try:
        df = pd.read_csv(input_file)
        
        click.echo(f"File contains {len(df)} rows and {len(df.columns)} columns")
        click.echo(f"Columns: {list(df.columns)}")
        
        # Check for common GPS column names
        common_lat_cols = ['latitude', 'lat', 'location-lat', 'Latitude']
        common_lon_cols = ['longitude', 'lon', 'location-long', 'Longitude']
        common_time_cols = ['datetime', 'timestamp', 'time', 'Date_Time']
        
        lat_col = next((col for col in common_lat_cols if col in df.columns), None)
        lon_col = next((col for col in common_lon_cols if col in df.columns), None)
        time_col = next((col for col in common_time_cols if col in df.columns), None)
        
        if lat_col:
            click.echo(f"✓ Found latitude column: {lat_col}")
        else:
            click.echo("✗ No latitude column detected")
            
        if lon_col:
            click.echo(f"✓ Found longitude column: {lon_col}")
        else:
            click.echo("✗ No longitude column detected")
            
        if time_col:
            click.echo(f"✓ Found datetime column: {time_col}")
        else:
            click.echo("✗ No datetime column detected")
        
        if lat_col and lon_col and time_col:
            click.echo("✓ File appears to be compatible with MagGeo")
            click.echo(f"Suggested command:")
            click.echo(f"maggeo annotate {input_file} --lat-col {lat_col} --lon-col {lon_col} --datetime-col {time_col}")
        else:
            click.echo("✗ File may need column name adjustments")
            
    except Exception as e:
        raise click.ClickException(f"Error validating file: {str(e)}")


@main.command()
def info():
    """Display MagGeo package information."""
    
    click.echo(f"MagGeo version: {__version__}")
    click.echo("A Python package for fusing GPS trajectories with geomagnetic data")
    click.echo("")
    click.echo("Authors: Fernando Benitez-Paez, Urška Demšar, Jed Long, Ciaran Beggan")
    click.echo("Citation: Benitez-Paez et al. (2021) Movement Ecology 9:31")
    click.echo("Repository: https://github.com/MagGeo/MagGeo")
    click.echo("")
    click.echo("Available commands:")
    click.echo("  annotate  - Annotate GPS trajectories with geomagnetic data")
    click.echo("  swarm     - Download and manage Swarm satellite data")
    click.echo("  validate  - Validate GPS file format")
    click.echo("  info      - Display this information")


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

setup(
    name='maggeo',
    version='0.2.0',
    author='Fernando Benitez-Paez, Urška Demšar, Jed Long, Ciaran Beggan',
    author_email='Fernando.Benitez@st-andrews.ac.uk',
    description='A Python library for annotating GPS trajectories with Swarm satellite geomagnetic data.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/MagGeo/MagGeo-Annotation-Program',
    packages=find_packages(),
    install_requires=[
        'numpy>=1.21.0',
        'pandas>=1.3.0',
        'scipy>=1.7.0',
        'matplotlib>=3.4.0',
        'viresclient>=0.11.0',
        'chaosmagpy>=0.11.0',
        'tqdm>=4.60.0',
        'requests>=2.25.0',
        'pyyaml>=5.4.0',
        'pooch>=1.3.0',
        'xarray>=0.18.0',
        'netcdf4>=1.5.0',
    ],
    extras_require={
        'dev': [
            'pytest>=6.0.0',
            'pytest-cov>=2.12.0',
            'black>=21.0.0',
            'flake8>=3.9.0',
            'mypy>=0.910',
        ],
        'notebooks': [
            'jupyter>=1.0.0',
            'jupyterlab>=3.0.0',  
            'ipywidgets>=7.6.0',
        ],
        'geo': [
            'cartopy>=0.19.0',
            'geopandas>=0.9.0',
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9', 
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Animal Movement Analytics',
        'Topic :: Scientific/Engineering :: Data Fusion',
        'Topic :: Scientific/Engineering :: Geomagnetism',
        'Topic :: Scientific/Engineering :: Data Science Ecology',
        'Topic :: Scientific/Engineering :: Animal tracking',
    ],
    python_requires='>=3.8',
    keywords='GPS, Swarm, satellite, geomagnetic, data fusion, animal tracking',
    entry_points={
        'console_scripts': [
            'maggeo=maggeo.core:main',
        ],
    },
)



import pandas as pd
import numpy as np
from .auxiliaryfunctions import gg_to_geo
from typing import Dict, Any
from chaosmagpy import load_CHAOS_matfile, data_utils
import pooch

# 3. This function computes the CHAOS ground values based on the GPS trajectory and the Swarm residuals.
# This is key part of the MagGeo workflow, as it allows to compute the magnetic field components at ground level or 
# bird level, using the CHAOS model and the Swarm residuals.
# See the paper: Benitez-Paez, F., Brum-Bastos, V.d., Beggan, C.D. et al. Fusion of wildlife tracking and satellite geomagnetic data for the study of animal migration. Mov Ecol 9, 31 (2021). https://doi.org/10.1186/s40462-021-00268-4


def chaos_ground_values(GPS_ResInt):

    """Compute the CHAOS ground values based on GPS trajectory and Swarm residuals.
    This function loads the CHAOS model, computes the geocentric coordinates, and calculates the magnetic field components at ground level.
    It transforms the magnetic field components from geocentric to geodetic frame and returns the results.
    
    Parameters
    ----------
    utilities_dir : str
        Directory where the utilities are located, including the CHAOS model.
    GPS_ResInt : pd.DataFrame
        DataFrame containing the GPS trajectory and Swarm residuals.

    Returns
    -------
    tuple
        Tuple containing the computed magnetic field components in the geodetic frame:
        (X_obs, Y_obs, Z_obs, X_obs_internal, Y_obs_internal, Z_obs_internal).
    
    """

    chaos_matfile = pooch.retrieve(
    #"http://www.spacecenter.dk/files/magnetic-models/CHAOS-7/CHAOS-7.15.mat",
    "http://www.spacecenter.dk/files/magnetic-models/CHAOS-8/_downloads/21c4a48c1467d2ad6f99621910f5399f/CHAOS-8.3.mat",
    known_hash="cd8b38fe88f6b75cfd3892ed224f125c1a73bd642ddd4f28e303345aa292e629",
    #progressbar=True,
    )

    # 1. Load the CHAOS model from the maggeo/models directory using a package-relative path
    #chaos_mat_path = importlib.resources.files('maggeo.models').joinpath('CHAOS-7.mat')
    # use chaos_mat_path (as a pathlib.Path object)
    model = load_CHAOS_matfile(str(chaos_matfile))
    theta = 90-GPS_ResInt['Latitude'].values
    phi = GPS_ResInt['Longitude'].values
    alt=GPS_ResInt['Altitude'].values
    rad_geoc_ground, theta_geoc_ground, sd_ground, cd_ground = gg_to_geo(alt, theta) 
    # gg_to_geo, will transform the coordinates from geocentric values to geodesic values. Altitude must be in km
    time = data_utils.mjd2000(
        pd.DatetimeIndex(GPS_ResInt['DateTime']).year,
        pd.DatetimeIndex(GPS_ResInt['DateTime']).month,
        pd.DatetimeIndex(GPS_ResInt['DateTime']).day
    )
    
    #2. Compute the core, crust and magnetosphere contributions at the altitude level.
    B_r_core, B_t_core, B_phi_core = model.synth_values_tdep(time, rad_geoc_ground, theta_geoc_ground, phi) #Core Contribution
    B_r_crust, B_t_crust, B_phi_crust = model.synth_values_static(rad_geoc_ground, theta_geoc_ground, phi) #Crust Contribution
    B_r_magneto, B_t_magneto, B_phi_magneto = model.synth_values_gsm(time, rad_geoc_ground, theta_geoc_ground, phi) #Magnetosphere contribution.

    #3. Change the direction of the axis from XYZ to r,theta and phi.
    B_r_swarm, B_t_swarm, B_phi_swarm = -GPS_ResInt['C_res'], -GPS_ResInt['N_res'], GPS_ResInt['E_res']

    #4. Compute the magnetic component (r,theta,phi) at ground level.
    B_r_ground = B_r_core + B_r_crust + B_r_magneto + B_r_swarm #(-Z)
    B_t_ground = B_t_core + B_t_crust + B_t_magneto + B_t_swarm #(-X)
    B_phi_ground = B_phi_core + B_phi_crust +B_phi_magneto + B_phi_swarm #(Y)

    #4b. Compute the CHAOS internal magnetic component (r,theta,phi) at ground level.
    B_r_ground_internal = B_r_core + B_r_crust #(-Z)
    B_t_ground_internal = B_t_core + B_t_crust #(-X)
    B_phi_ground_internal = B_phi_core + B_phi_crust #(Y)

    #5. Convert B_r_, B_t_, and B_phi to XYZ (NEC)
    Z_chaos = -B_r_ground   #Z
    X_chaos = -B_t_ground   #X
    Y_chaos = B_phi_ground  #Y

    #5b. Convert full field with CHAOS internal only B_r_, B_t_, and B_phi to XYZ (NEC)
    Z_chaos_internal = -B_r_ground_internal #Z
    X_chaos_internal = -B_t_ground_internal #X
    Y_chaos_internal = B_phi_ground_internal #Y

    #6. Rotate the X(N) and Z(C) magnetic field values of the chaos models into the geodetic frame using the sd and cd (sine and cosine d from gg_to_geo) 
    X_obs = X_chaos*cd_ground + Z_chaos*sd_ground #New N
    Z_obs = Z_chaos*cd_ground - X_chaos*sd_ground #New C
    Y_obs = Y_chaos # New E

    #6b. Rotate the X(N) and Z(C) magnetic field values of the internal part of chaos models into the geodetic frame using the sd and cd (sine and cosine d from gg_to_geo) 
    X_obs_internal = X_chaos_internal *cd_ground + Z_chaos_internal*sd_ground #New N
    Z_obs_internal = Z_chaos_internal *cd_ground - X_chaos_internal*sd_ground #New C
    Y_obs_internal = Y_chaos_internal # New E

    return X_obs, Y_obs, Z_obs, X_obs_internal, Y_obs_internal, Z_obs_internal



    """
    Process a single chunk of GPS data for CHAOS calculations.
    
    This is a helper function for parallel processing.
    
    Parameters
    ----------
    gps_resint_chunk : pd.DataFrame
        Chunk of GPS data with interpolated residuals
        
    Returns
    -------
    pd.DataFrame
        DataFrame chunk with CHAOS calculations
    """
    # Calculate CHAOS ground values for this chunk
    X_obs, Y_obs, Z_obs, X_obs_internal, Y_obs_internal, Z_obs_internal = chaos_ground_values(gps_resint_chunk)
    
    # Create a copy of the chunk and add CHAOS results
    result_chunk = gps_resint_chunk.copy()
    result_chunk['N'] = X_obs
    result_chunk['E'] = Y_obs
    result_chunk['C'] = Z_obs
    result_chunk['N_Obs'] = X_obs_internal
    result_chunk['E_Obs'] = Y_obs_internal
    result_chunk['C_Obs'] = Z_obs_internal
    
    # Calculate additional geomagnetic components
    result_chunk['H'] = np.sqrt((result_chunk['N']**2 + result_chunk['E']**2))
    result_chunk['D'] = np.degrees(np.arctan2(result_chunk['E'], result_chunk['N']))
    result_chunk['I'] = np.degrees(np.arctan2(result_chunk['C'], result_chunk['H']))
    result_chunk['F'] = np.sqrt((result_chunk['N']**2 + result_chunk['E']**2 + result_chunk['C']**2))
    
    return result_chunk